from big_thing_py.core.request import *
from big_thing_py.super.super_mqtt_message import *
from big_thing_py.core.function import *
from func_timeout import FunctionTimedOut


class MXSuperScheduleRequest(MXRequest):
    def __init__(self, trigger_msg: MXSuperScheduleMessage = None, result_msg: MXSuperScheduleResultMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.SUPER_SCHEDULE

        self.trigger_msg: MXSuperScheduleMessage
        self.result_msg: MXSuperScheduleResultMessage
        self._check_duration: float = 0.0
        self._confirm_duration: float = 0.0
        self._status: MXScheduleStatus = MXScheduleStatus.INIT
        self._running: bool = False

        # sub_request_key = make_sub_request_key(subschedule_result_msg.sub_service_name, subschedule_result_msg.sub_service_request_order)
        # sub_request_key: {sub_service_name}@{sub_service_request_order}
        self._sub_service_request_table: Dict[str, MXSubServiceRequest] = {}

    def set_result_msg(self, result_msg: MXSuperScheduleResultMessage):
        self.result_msg = result_msg

    def get_result_msg(self):
        return self.result_msg

    def is_completed(self):
        if self.result_msg:
            return True
        else:
            return False

    def check_duration(self):
        return self._check_duration

    def confirm_duration(self):
        return self._confirm_duration

    def put_sub_service_schedule_result_msg(
        self, sub_service_request_key: str, result_msg: MXSubScheduleResultMessage, timeout: float = None
    ) -> bool:
        if sub_service_request_key not in self._sub_service_request_table:
            MXLOG_DEBUG(
                f'Could not find key {sub_service_request_key} in sub_service_request_table... This key is not mine!',
                'yellow',
            )
            return False

        target_sub_service_request = self._sub_service_request_table[sub_service_request_key]
        for candidate_request in target_sub_service_request._candidate_request_list:
            thing_check = candidate_request.trigger_msg.target_thing_name == result_msg.target_thing_name
            middleware_check = candidate_request.trigger_msg.target_middleware_name == result_msg.target_middleware_name
            request_ID_check = candidate_request.trigger_msg.request_ID == result_msg.request_ID
            status_check = (result_msg.status == MXScheduleStatus.CONFIRM and candidate_request._status == MXScheduleStatus.SELECT) or (
                result_msg.status == MXScheduleStatus.CHECK
            )

            if thing_check and middleware_check and request_ID_check and status_check:
                candidate_request.put_result_msg(result_msg, timeout=timeout)
                return True
        else:
            raise Exception(f'No sub_service {sub_service_request_key} found in _sub_service_request_table')

    def generate_sub_service_schedule_request(
        self,
        sub_service_request_list: List['MXSubServiceRequest'],
        hierarchical_function_service_table: List[MXFunction],
    ):
        for sub_service_request in sub_service_request_list:
            target_sub_service_request = MXSubServiceRequest(
                sub_service_type=sub_service_request._sub_service_type,
                sub_service_request_order=sub_service_request._sub_service_request_order,
            )
            # TODO: 해당 로직 부분 분리하기
            # 미들웨어 마다 sub_service_request가 실행하고자 하는 sub_service이 존재하는 지 체크한다.
            target_sub_service = target_sub_service_request._sub_service_type

            candidate_middleware_name_list = []
            for function_service in hierarchical_function_service_table:
                name_check = target_sub_service.name == function_service.name
                if name_check and function_service.middleware_name not in candidate_middleware_name_list:
                    candidate_middleware_name_list.append(function_service.middleware_name)
                    target_sub_service.middleware_name = function_service.middleware_name

            candidate_request_list = []
            for middleware_name in candidate_middleware_name_list:
                sub_service_schedule_msg = MXSubScheduleMessage(
                    sub_service_name=target_sub_service.name,
                    target_thing_name='SUPER',
                    target_middleware_name=middleware_name,
                    requester_middleware_name=self.trigger_msg.requester_middleware_name,
                    super_thing_name=self.trigger_msg.super_thing_name,
                    super_service_name=self.trigger_msg.super_service_name,
                    sub_service_request_order=target_sub_service_request._sub_service_request_order,
                    scenario=self.trigger_msg.scenario,
                    period=self.trigger_msg.period,
                    tag_list=[dict(name=tag.name) for tag in target_sub_service.tag_list],
                    range_type=target_sub_service.range_type,
                    request_ID=None,
                    status=MXScheduleStatus.INIT,
                )

                # NOTE:sub_service_request에 대한 부분이 미들웨어당 1개만 생성될 것으로 예상됨. len(candidate_request_list) == 1 ?
                # TODO: _result_msg도 생성해서 넣어서 나중에 결과 토픽을 구독할 때 topic()으로 간단하게 사용할 수 있게 하면 좋을 것 같다.
                sub_service_schedule_request = MXSubScheduleRequest(trigger_msg=sub_service_schedule_msg)
                candidate_request_list.append(sub_service_schedule_request)

            target_sub_service_request._candidate_request_list = candidate_request_list
            sub_service_request_key = make_sub_service_request_key(
                sub_service_name=target_sub_service.name,
                sub_service_request_order=target_sub_service_request._sub_service_request_order,
            )
            self._sub_service_request_table[sub_service_request_key] = target_sub_service_request

            if len(target_sub_service_request._candidate_request_list) == 0:
                for function_service in hierarchical_function_service_table:
                    MXLOG_DEBUG(f'sub_service found! - {function_service.name}|{function_service.thing_name}|{function_service.middleware_name}')
                raise Exception(
                    f'No candidate sub_service found in key:{sub_service_request_key} {self.trigger_msg.super_service_name} super service'
                )


class MXSuperExecuteRequest(MXRequest):
    def __init__(self, trigger_msg: MXSuperExecuteMessage = None, super_schedule_request: MXSuperScheduleRequest = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=None)
        self._action_type = MXActionType.SUPER_EXECUTE

        self.trigger_msg: MXSuperExecuteMessage
        self.result_msg: MXSuperExecuteResultMessage
        self._running: bool = False

        # scheduling이 끝나면 해당 scenario에 대해 어떤 sub_service_type에 대해 thing을 선택할 것인지 정해진다.
        # sub_request_key = make_sub_request_key(subschedule_result_msg.sub_service_name, subschedule_result_msg.sub_service_request_order)
        # sub_request_key: {sub_service_name}@{sub_service_request_order}
        self._sub_service_request_table: Dict[str, MXSubServiceRequest] = {}

        if super_schedule_request:
            self.generate_sub_service_execute_request_list(super_schedule_request=super_schedule_request)

    def set_result_msg(self, result_msg: MXSuperExecuteResultMessage):
        self.result_msg = result_msg

    def get_result_msg(self):
        return self.result_msg

    def is_completed(self):
        if self.result_msg:
            return True
        else:
            return False

    def put_sub_service_execute_result_msg(self, sub_service_request_key: str, result_msg: MXSubExecuteResultMessage, timeout: float = None) -> bool:
        if sub_service_request_key not in self._sub_service_request_table:
            MXLOG_DEBUG(
                f'Could not find key {sub_service_request_key} in sub_service_request_table... This key is not mine!',
                'yellow',
            )
            return False

        target_sub_service_request = self._sub_service_request_table[sub_service_request_key]
        for target_request in target_sub_service_request._target_request_list:
            target_thing_check = target_request.trigger_msg.target_thing_name == result_msg.target_thing_name
            target_middleware_check = target_request.trigger_msg.target_middleware_name == result_msg.target_middleware_name
            request_ID_check = target_request.trigger_msg.request_ID == result_msg.request_ID
            if target_thing_check and target_middleware_check and request_ID_check:
                target_request.put_result_msg(result_msg, timeout=timeout)
                return True
        else:
            raise Exception(f'No sub_service {sub_service_request_key} found in _sub_service_request_table')

    def get_sub_service_execute_result_msg_list(self, sub_service_request_key: str, timeout: float = None) -> List[MXSubExecuteResultMessage]:
        result_msg_list = []
        target_sub_service_request = self._sub_service_request_table[sub_service_request_key]
        for target_request in target_sub_service_request._target_request_list:
            result_msg_list.append(target_request.get_result_msg(timeout=timeout))
        return result_msg_list

    def generate_sub_service_execute_request_list(self, super_schedule_request: 'MXSuperScheduleRequest' = None) -> 'MXSuperExecuteRequest':
        # schedule 단계에서 MXSuperExecuteRequest를 생성한다.
        for sub_service_request_key, sub_service_request in super_schedule_request._sub_service_request_table.items():
            self._sub_service_request_table[sub_service_request_key] = MXSubServiceRequest(
                sub_service_type=sub_service_request._sub_service_type,
                sub_service_request_order=sub_service_request._sub_service_request_order,
                candidate_request_list=sub_service_request._candidate_request_list,
            )
            target_request_list = []
            for candidate_request in sub_service_request._candidate_request_list:
                if candidate_request._status == MXScheduleStatus.CONFIRM:
                    target_request_list.append(to_sub_service_execute_request(candidate_request))
            self._sub_service_request_table[sub_service_request_key]._target_request_list = target_request_list
        return self


class MXSubScheduleRequest(MXRequest):
    def __init__(self, trigger_msg: MXSubScheduleMessage = None, result_msg: MXSubScheduleResultMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.SUB_SCHEDULE

        self.trigger_msg: MXSubScheduleMessage
        self.result_msg: MXSubScheduleResultMessage
        self._status: MXScheduleStatus = MXScheduleStatus.INIT
        self._result_queue = Queue()

    def __getstate__(self):
        state = self.__dict__.copy()

        del state['_result_queue']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self._result_queue = Queue()

    def put_result_msg(self, result_msg: MXSubScheduleResultMessage, timeout: float = None):
        self._result_queue.put(result_msg, timeout=timeout)

    def get_result_msg(self, timeout: float = None):
        try:
            return self._result_queue.get(timeout=timeout)
        except Empty:
            raise FunctionTimedOut


class MXSubExecuteRequest(MXRequest):
    def __init__(self, trigger_msg: MXSubExecuteMessage = None, result_msg: MXSubExecuteResultMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.SUB_EXECUTE

        self.trigger_msg: MXSubExecuteMessage
        self.result_msg: MXSubExecuteResultMessage
        self._result_queue = Queue()

    def __getstate__(self):
        state = self.__dict__.copy()

        del state['_result_queue']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self._result_queue = Queue()

    def put_result_msg(self, result_msg: MXSubExecuteResultMessage, timeout: float = None):
        self._result_queue.put(result_msg, timeout=timeout)

    def get_result_msg(self, timeout: float = None) -> MXSubExecuteResultMessage:
        try:
            return self._result_queue.get(timeout=timeout)
        except Empty:
            raise FunctionTimedOut


class MXSubServiceRequest:
    def __init__(
        self,
        sub_service_type: MXFunction,
        candidate_request_list: List[MXSubScheduleRequest] = [],
        target_request_list: List[MXSubExecuteRequest] = [],
        sub_service_request_order: int = None,
    ) -> None:
        self._sub_service_type = sub_service_type
        self._candidate_request_list = candidate_request_list
        self._target_request_list = target_request_list
        self._sub_service_request_order = sub_service_request_order

    def __eq__(self, o: 'MXSubServiceRequest') -> bool:
        instance_check = isinstance(o, MXSubServiceRequest)
        # sub_service_type_check = (o._sub_service_type == self._sub_service_type)

        def sub_service_type_check():
            return_type_check = o._sub_service_type._return_type == self._sub_service_type._return_type
            exec_time_check = o._sub_service_type._exec_time == self._sub_service_type._exec_time
            timeout_check = o._sub_service_type._timeout == self._sub_service_type._timeout
            range_type_check = o._sub_service_type._range_type == self._sub_service_type._range_type

            if return_type_check and exec_time_check and timeout_check and range_type_check:
                return True
            else:
                return False

        candidate_request_list_check = o._candidate_request_list == self._candidate_request_list
        target_request_list_check = o._target_request_list == self._target_request_list
        sub_service_request_order_check = o._sub_service_request_order == self._sub_service_request_order

        return (
            instance_check
            and sub_service_type_check
            and candidate_request_list_check
            and target_request_list_check
            and sub_service_request_order_check
        )

    def __deepcopy__(self, memodict={}):
        new_instance = MXSubServiceRequest(self._sub_service_type, self._sub_service_request_order, list())
        new_instance.__dict__.update(self.__dict__)
        new_instance._sub_service_type = copy.deepcopy(self._sub_service_type, memodict)
        new_instance._sub_service_request_order = copy.deepcopy(self._sub_service_request_order, memodict)
        new_instance._candidate_request_list = list()
        new_instance._target_request_list = list()
        return new_instance


def to_sub_service_execute_request(sub_service_schedule_request: MXSubScheduleRequest) -> MXSubExecuteRequest:
    sub_service_schedule_msg: MXSubScheduleMessage = sub_service_schedule_request.trigger_msg
    sub_service_execute_msg = MXSubExecuteMessage(
        sub_service_schedule_msg.sub_service_name,
        'SUPER',
        sub_service_schedule_msg.target_middleware_name,
        sub_service_schedule_msg.requester_middleware_name,
        sub_service_schedule_msg.super_thing_name,
        sub_service_schedule_msg.super_service_name,
        sub_service_schedule_msg.sub_service_request_order,
        sub_service_schedule_msg.scenario,
        request_ID=sub_service_schedule_msg.request_ID,
    )
    sub_service_execute_request = MXSubExecuteRequest(trigger_msg=sub_service_execute_msg)
    return sub_service_execute_request


def to_sub_service_schedule_request(sub_service_execute_request: MXSubExecuteRequest) -> MXSubScheduleRequest:
    sub_service_execute_msg: MXSubExecuteMessage = sub_service_execute_request.trigger_msg
    sub_service_schedule_msg = MXSubScheduleMessage(
        sub_service_name=sub_service_execute_msg.sub_service_name,
        target_thing_name=sub_service_execute_msg.target_thing_name,
        target_middleware_name=sub_service_execute_msg.target_middleware_name,
        requester_middleware_name=sub_service_execute_msg.requester_middleware_name,
        super_thing_name=sub_service_execute_msg.super_thing_name,
        super_service_name=sub_service_execute_msg.super_service_name,
        sub_service_request_order=sub_service_execute_msg.sub_service_request_order,
        scenario=sub_service_execute_msg.scenario,
        request_ID=sub_service_execute_msg.request_ID,
    )
    sub_service_schedule_request = MXSubScheduleRequest(trigger_msg=sub_service_schedule_msg)
    return sub_service_schedule_request



from big_thing_py.utils.common_util import *
from big_thing_py.utils.exception_util import *
from big_thing_py.utils.log_util import *


def read_file(path: str) -> List[str]:
    try:
        with open(path, 'r') as f:
            return f.readlines()
    except FileNotFoundError as e:
        print_error(e)
        MXLOG_DEBUG(f'File not found: {path}')
        raise e


def write_file(path: str, content: Union[str, List[str]]) -> None:
    try:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            if isinstance(content, str):
                f.write(content)
            elif isinstance(content, list):
                f.writelines(content)
        return path
    except FileNotFoundError as e:
        print_error(e)
        MXLOG_DEBUG(f'Path not found: {path}')
        raise e



from big_thing_py.utils import *

from enum import Enum, auto


try:
    from RF24 import RF24, RF24_PA_LOW, RF24_PA_MAX
except ImportError:
    pass
    # print('RF24 not found', 'red')


class MXRFProtocol(Enum):
    REG = 'REG'
    RACK = 'RACK'
    VAL = 'VAL'
    LIVE = 'LIVE'
    EXEC = 'EXEC'
    EACK = 'EACK'


class MXRFPowerMode(Enum):
    try:
        LOW = RF24_PA_LOW
        HIGH = RF24_PA_MAX
    except Exception:
        pass
        # MXLOG_DEBUG('RF24_PA_LOW or RF24_PA_MAX not found', 'red')


class MXRFMessage:
    def __init__(
        self, protocol_type: MXRFProtocol, device_id: bytearray, service_name: str, payload: bytearray
    ) -> None:
        self.protocol_type = protocol_type
        self.device_id = device_id
        self.service_name = service_name
        self.payload = payload


def decode_RF_message(raw_msg: bytearray) -> MXRFMessage:
    try:
        protocol_type = MXRFProtocol(raw_msg[:4].rstrip().rstrip(b'\x00').decode().upper())
        device_id = raw_msg[4:8].decode()
        service_name = raw_msg[8:16].rstrip().rstrip(b'\x00').decode()
        payload = raw_msg[16:32].decode()

        return MXRFMessage(protocol_type=protocol_type, device_id=device_id, service_name=service_name, payload=payload)
    except Exception as e:
        MXLOG_DEBUG('[mqtt_util.py|dict_to_json_string]' + str(e), 'red')
        return False


def encode_RF_message(msg: MXRFMessage) -> bytearray:
    protocol_type = bytearray(msg.protocol_type.value.encode())
    device_id = msg.device_id
    service_name = bytearray(msg.service_name.encode())
    payload = msg.payload

    return protocol_type + device_id + service_name + payload



from .common_util import *
from .file_util import *
from .json_util import *
from .log_util import *
from .mqtt_util import *
from .rf_util import *
from .exception_util import *
from .api_util import *
from .base64_util import *



from big_thing_py.common.mxtype import *
from big_thing_py.core.mqtt_message import *


class MXRequest(metaclass=ABCMeta):
    def __init__(self, trigger_msg: MXMQTTMessage = None, result_msg: MXMQTTMessage = None) -> None:
        self._action_type: MXActionType = None
        self._trigger_msg = trigger_msg
        self._result_msg = result_msg

        # seconds
        self._duration: float = 0

    def duration(self):
        return self._duration

    def timer_start(self):
        self._trigger_msg.set_timestamp(get_current_datetime())

    def timer_end(self):
        try:
            self._result_msg.set_timestamp(get_current_datetime())
            self._duration = self._result_msg.timestamp - self._trigger_msg.timestamp
        except Exception:
            self._duration = time.time() - self._trigger_msg.timestamp
        return self.duration()

    @property
    def result_msg(self):
        return self._result_msg

    @result_msg.setter
    def result_msg(self, result: MXMQTTReceiveMessage):
        self._result_msg = result

    @property
    def trigger_msg(self):
        return self._trigger_msg

    @trigger_msg.setter
    def trigger_msg(self, result: MXMQTTSendMessage):
        self._trigger_msg = result

    @property
    def action_type(self):
        return self._action_type


class MXRegisterRequest(MXRequest):
    def __init__(self, trigger_msg: MXMQTTMessage = None, result_msg: MXMQTTMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.REGISTER

        self._trigger_msg: MXExecuteMessage
        self._result_msg: MXExecuteResultMessage


class MXExecuteRequest(MXRequest):
    def __init__(self, trigger_msg: MXExecuteMessage = None, result_msg: MXExecuteResultMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.EXECUTE

        self._trigger_msg: MXExecuteMessage
        self._result_msg: MXExecuteResultMessage

    @property
    def result_msg(self):
        return self._result_msg

    @result_msg.setter
    def result_msg(self, result: MXExecuteResultMessage):
        self._result_msg = result

    @property
    def trigger_msg(self):
        return self._trigger_msg

    @trigger_msg.setter
    def trigger_msg(self, result: MXExecuteMessage):
        self._trigger_msg = result


class MXInnerExecuteRequest(MXRequest):
    def __init__(self, trigger_msg: MXExecuteMessage = None, result_msg: MXExecuteResultMessage = None) -> None:
        super().__init__(trigger_msg=trigger_msg, result_msg=result_msg)
        self._action_type = MXActionType.INNER_EXECUTE

        self._trigger_msg: MXExecuteMessage
        self._result_msg: MXExecuteResultMessage

    @property
    def result_msg(self):
        return self._result_msg

    @result_msg.setter
    def result_msg(self, result: MXExecuteResultMessage):
        self._result_msg = result

    @property
    def trigger_msg(self):
        return self._trigger_msg

    @trigger_msg.setter
    def trigger_msg(self, result: MXExecuteMessage):
        self._trigger_msg = result



from big_thing_py.utils.exception_util import *
from big_thing_py.utils.log_util import *
from big_thing_py.utils.json_util import *

from gmqtt import Message as MQTTMessage


def encode_MQTT_message(topic: str, payload: Union[str, dict]) -> MQTTMessage:
    try:
        if isinstance(payload, str):
            payload = bytes(payload, encoding='utf-8')
        elif isinstance(payload, dict):
            payload = dict_to_json_string(payload)

        return MQTTMessage(topic, payload)
    except Exception as e:
        print_error(e)
        raise e


def decode_MQTT_message(msg: MQTTMessage, mode=dict) -> Tuple[str, dict]:
    topic = msg.topic
    payload = msg.payload

    if isinstance(topic, bytes):
        topic = topic.decode()
    if isinstance(payload, bytes):
        payload = payload.decode()

    if isinstance(payload, str):
        if mode == str:
            return topic, payload
        elif mode == dict:
            return topic, json_string_to_dict(payload)
        else:
            raise MXNotSupportedError(f'Unexpected mode!!! - {mode}')
    elif isinstance(payload, dict):
        if mode == str:
            return topic, dict_to_json_string(payload)
        elif mode == dict:
            return topic, payload
        else:
            raise MXNotSupportedError(f'Unexpected mode!!! - {mode}')
    else:
        raise MXNotSupportedError(f'Unexpected type!!! - {type(payload)}')


def topic_split(topic: str):
    return topic.split('/')


def topic_join(topic: List[str]):
    return '/'.join(topic)


def unpack_mqtt_message(msg: MQTTMessage) -> Tuple[List[str], str]:
    topic, payload = decode_MQTT_message(msg, dict)
    topic = topic_split(topic)

    return topic, payload


def pack_mqtt_message(topic_list: List[str], payload: str) -> MQTTMessage:
    topic = topic_join(topic_list)
    msg = encode_MQTT_message(topic, payload)

    return msg



from big_thing_py.utils import *


class MXTag:
    def __init__(self, name: str):
        self._name: str = name

        if check_valid_identifier(self._name) == MXErrorCode.INVALID_DATA:
            raise MXValueError(f'name cannot be empty & can only contain alphanumeric characters and underscores. name: {self._name}')
        elif check_valid_identifier(self._name) == MXErrorCode.TOO_LONG_IDENTIFIER:
            raise MXValueError(f'too long identifier. name: {self._name}, length: {len(self._name)}')

    def __eq__(self, o: 'MXTag') -> bool:
        name_check = o._name == self._name

        return isinstance(o, MXTag) and name_check

    def __str__(self) -> str:
        return self._name

    def dict(self):
        return {'name': self._name}

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def name(self):
        return self._name

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @name.setter
    def name(self, name):
        self._name = name



from big_thing_py.core.tag import *
from big_thing_py.core.service_model import SkillValue, SkillFunction
from abc import *
from asyncio import iscoroutinefunction


class MXService(metaclass=ABCMeta):

    def __init__(
        self,
        name: str,
        category: Union[SkillValue, SkillFunction],
        func: Union[Callable, Coroutine],
        tag_list: List[MXTag],
        energy: float,
        desc: str,
        thing_name: str,
        middleware_name: str,
    ) -> None:
        self._name = name
        self._category = category
        self._func = func
        self._tag_list = tag_list
        self._energy = energy
        self._desc = desc
        self._thing_name = thing_name
        self._middleware_name = middleware_name

        if not callable(self._func) and not iscoroutinefunction(self._func):
            raise MXValueError(f'func must be callable or coroutine function.')
        if not issubclass(self._category, SkillValue) and not issubclass(self._category, SkillFunction):
            raise MXValueError(f'category must be SkillValue or SkillFunction class.')
        if not self._name:
            self._name = self._category.function_id if issubclass(self._category, SkillFunction) else self._category.value_id
        if check_valid_identifier(self._name) == MXErrorCode.INVALID_DATA:
            raise MXValueError(f'name cannot be empty & can only contain alphanumeric characters and underscores. name: {self._name}')
        elif check_valid_identifier(self._name) == MXErrorCode.TOO_LONG_IDENTIFIER:
            raise MXValueError(f'too long identifier. name: {self._name}, length: {len(self._name)}')

        if any([not isinstance(tag, MXTag) for tag in self._tag_list]):
            raise MXValueError('tag_list must contain MXTag object')
        else:
            self.sort_tag()

        if not isinstance(self._energy, (int, float)):
            raise MXValueError(f'energy must be int or float')
        if not isinstance(self._desc, str):
            raise MXValueError(f'desc must be str')
        if not isinstance(self._thing_name, str):
            raise MXValueError(f'thing_name must be str')
        if not isinstance(self._middleware_name, str):
            raise MXValueError(f'middleware_name must be str')

        # NOTE (thsvkd): test new service model, SkillValue, SkillFunction
        # if self._category == MXFunctionCategory.undefined:
        #     self._category = MXFunctionCategory.get(self._name)

    def __eq__(self, o: 'MXService') -> bool:
        instance_check = isinstance(o, MXService)
        name_check = o._name == self._name
        # thing_name_check = o._thing_name == self._thing_name
        # middleware_name_check = o._middleware_name == self._middleware_name
        tag_list_check = o._tag_list == self._tag_list
        # func_check = o._func == self._func
        energy_check = o._energy == self._energy

        # return instance_check and name_check and thing_name_check and middleware_name_check and tag_list_check and func_check and energy_check
        return instance_check and name_check and tag_list_check and energy_check

    def __getstate__(self):
        state = self.__dict__.copy()

        del state['_func']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self._func = None

    def add_tag(self, tag: Union[MXTag, List[MXTag]]) -> 'MXService':
        if not isinstance(tag, (MXTag, list)):
            raise Exception('tag must be MXTag object')

        if isinstance(tag, MXTag):
            if not tag in self._tag_list:
                self._tag_list.append(tag)
        elif all(isinstance(t, MXTag) for t in tag):
            for t in tag:
                if not t in self._tag_list:
                    self._tag_list.append(t)
        self._tag_list = sorted(self._tag_list, key=lambda x: x.name)

        return self

    def remove_tag(self, tag: str) -> 'MXService':
        if not isinstance(tag, str):
            raise Exception('tag to remove must be str')

        for t in self._tag_list:
            if t.name == tag:
                self._tag_list.remove(MXTag(tag))

    def sort_tag(self) -> 'MXService':
        self._tag_list = sorted(self._tag_list, key=lambda x: x.name)
        return self

    @abstractmethod
    def dict(self) -> dict:
        pass

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def name(self) -> str:
        return self._name

    @property
    def category(self) -> Union[SkillValue, SkillFunction]:
        return self._category

    @property
    def thing_name(self) -> str:
        return self._thing_name

    @property
    def middleware_name(self) -> str:
        return self._middleware_name

    @property
    def tag_list(self) -> List[MXTag]:
        return self._tag_list

    @property
    def desc(self) -> str:
        return self._desc

    @property
    def func(self) -> Callable:
        return self._func

    @property
    def energy(self) -> float:
        return self._energy

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @category.setter
    def category(self, category: Union[SkillValue, SkillFunction]) -> None:
        self._category = category

    @thing_name.setter
    def thing_name(self, thing_name: str) -> None:
        self._thing_name = thing_name

    @middleware_name.setter
    def middleware_name(self, middleware_name: str) -> None:
        self._middleware_name = middleware_name

    @tag_list.setter
    def tag_list(self, tag_list: List[MXTag]) -> None:
        self._tag_list = tag_list

    @desc.setter
    def desc(self, desc: str) -> None:
        self._desc = desc

    @func.setter
    def func(self, func: Callable) -> None:
        self._func = func

    @energy.setter
    def energy(self, energy: float) -> None:
        self._energy = energy



from big_thing_py.common import *
from big_thing_py.utils.log_util import *
from big_thing_py.utils.exception_util import *

import json


def json_string_to_dict(json_string: str) -> Union[str, dict, list]:
    try:
        if type(json_string) not in [str, bytes]:
            raise json.JSONDecodeError('input string must be str, bytes type.', json_string, 0)
        else:
            return json.loads(json_string)

    except json.JSONDecodeError as e:
        MXLOG_DEBUG(f'[json_string_to_dict] input string must be json format string. return raw object... ', 'red')
        print_error(e)
        return json_string


def dict_to_json_string(dict_object: Union[dict, list, str], ensure_ascii: bool = True, pretty: bool = True, indent: int = 4) -> Union[str, bool]:
    try:
        if type(dict_object) == dict:
            if pretty:
                return json.dumps(dict_object, ensure_ascii=ensure_ascii, sort_keys=True, indent=indent)
            else:
                return json.dumps(dict_object, ensure_ascii=ensure_ascii)
        elif type(dict_object) == list:
            if pretty:
                return '\n'.join([json.dumps(item, ensure_ascii=ensure_ascii, sort_keys=True, indent=indent) for item in dict_object])
            else:
                return '\n'.join([json.dumps(item, ensure_ascii=ensure_ascii) for item in dict_object])
        else:
            if pretty:
                json.dumps(json.loads(dict_object), ensure_ascii=ensure_ascii, sort_keys=True, indent=indent)
            else:
                return str(dict_object)
    except Exception as e:
        MXLOG_DEBUG(f'[dict_to_json_string] input object must be dict or list or str. return False... ', 'red')
        print_error(e)
        return False



from big_thing_py.core.function import MXFunction
from big_thing_py.core.value import MXValue
from big_thing_py.core.service import MXService
from big_thing_py.core.tag import MXTag
from big_thing_py.core.mqtt_message import MXRegisterMessage, MXUnregisterMessage, MXAliveMessage, MXGetHomeMessage
from big_thing_py.core.device_model import Objects as DeviceCategory
from big_thing_py.utils import *


class MXThing:
    DEFAULT_NAME = 'default_big_thing'

    def __init__(
        self,
        name: str,
        nick_name: str,
        category: DeviceCategory,
        service_list: List[MXService],
        alive_cycle: float,
        is_super: bool,
        is_parallel: bool,
        is_builtin: bool = False,
        is_manager: bool = False,
        is_matter: bool = False,
        desc: str = '',
        version: str = sdk_version(),
        middleware_name: str = '',
    ):
        # base info
        self._name = name
        self._nick_name = nick_name
        self._category = category
        self._desc = desc
        self._version = version
        self._service_list = service_list
        self._alive_cycle = alive_cycle
        self._is_super = is_super
        self._is_parallel = is_parallel
        self._is_builtin = is_builtin
        self._is_manager = is_manager
        self._is_matter = is_matter
        self._middleware_name = middleware_name
        self._request_client_id = ''

        self._last_alive_time = 0
        self._subscribed_topic_set: Set[str] = set()
        self._dependency_table: Dict[str, List[MXValue]] = {}

        for service in self.service_list:
            self.add_service(service)

        if not self._name:
            self._name = self._category.name

        if check_valid_identifier(self._name) == MXErrorCode.INVALID_DATA:
            raise MXValueError(f'name cannot be empty & can only contain alphanumeric characters and underscores. name: {self._name}')
        elif check_valid_identifier(self._name) == MXErrorCode.TOO_LONG_IDENTIFIER:
            raise MXValueError(f'too long identifier. name: {self._name}, length: {len(self._name)}')

        if self._alive_cycle <= 0:
            raise MXValueError(f'alive cycle must be greater than 0')

    def __eq__(self, o: 'MXThing') -> bool:
        instance_check = isinstance(o, MXThing)
        name_check = o._name == self._name
        service_list_check = (o.function_list == self.function_list) and (o.value_list == self.value_list)
        alive_cycle_check = o._alive_cycle == self._alive_cycle
        is_super_check = o.is_super == self.is_super
        is_parallel_check = o.is_parallel == self.is_parallel

        return instance_check and name_check and service_list_check and alive_cycle_check and is_super_check and is_parallel_check

    def __getstate__(self):
        state = self.__dict__.copy()

        del state['_last_alive_time']
        del state['_subscribed_topic_set']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self._last_alive_time = 0
        self._subscribed_topic_set = set()

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def get_value(self, value_name: str) -> MXValue:
        for value in self.value_list:
            if value.name == value_name:
                return value

    def get_function(self, function_name: str) -> MXFunction:
        for function in self.function_list:
            if function.name == function_name:
                return function

    def add_service(self, service: MXService) -> 'MXThing':
        service.add_tag(MXTag(name=self._name))
        service.add_tag(MXTag(name=self._category.name))
        service.thing_name = self._name

        if isinstance(service, MXValue):
            value_getter_function = MXFunction(
                name=f'__{service.name}',
                func=service.func,
                return_type=service.type,
                category=service.category,
                tag_list=service.tag_list,
                energy=service.energy,
                desc=service.desc,
                thing_name=service.thing_name,
                middleware_name=service.middleware_name,
            )

            if not service in self.value_list:
                self._service_list.append(service)
            if not value_getter_function in self.function_list:
                self._service_list.append(value_getter_function)
        elif isinstance(service, MXFunction):
            if not service in self.function_list:
                self._service_list.append(service)
        else:
            raise MXTypeError(f'service_list must be list of MXFunction or MXValue object')

        return self

    def add_dependency(
        self,
        function: Union[MXFunction, str],
        value: Union[MXValue, str],
        set_value: Union[int, float, bool, str] = None,
        set_value_as_result: bool = False,
        set_value_with_cond: Callable = None,
    ) -> None:
        if not isinstance(function, MXFunction):
            function = self.get_function(function)
            if not function:
                raise MXValueError(f'function not found: {function}')

        if not isinstance(value, MXValue):
            value = self.get_value(value)
            if not value:
                raise MXValueError(f'value not found: {value}')

        dependency_info = dict(
            value=value,
            set_value=set_value,
            set_value_as_result=set_value_as_result,
            set_value_with_cond=set_value_with_cond,
        )
        if not function.name in self.dependency_table:
            self._dependency_table[function.name] = [dependency_info]
        else:
            self._dependency_table[function.name].append(dependency_info)

    def dict(self) -> dict:
        return {
            'name': self._name,
            'nick_name': self._nick_name,
            'category': self._category.name,
            'description': self._desc,
            'version': self._version,
            'alive_cycle': self._alive_cycle,
            'is_super': self.is_super,
            'is_parallel': self.is_parallel,
            'is_builtin': self.is_builtin,
            'is_manager': self._is_manager,
            'is_matter': self._is_matter,
            'client_id': self._request_client_id,
            'values': [value.dict() for value in self.value_list],
            'functions': [function.dict() for function in self.function_list],
        }

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def function_list(self) -> List[MXFunction]:
        return sorted([service for service in self._service_list if isinstance(service, MXFunction)], key=lambda x: x.name)

    @property
    def value_list(self) -> List[MXValue]:
        return sorted([service for service in self._service_list if isinstance(service, MXValue)], key=lambda x: x.name)

    @property
    def service_list(self) -> List[MXService]:
        return sorted(self._service_list, key=lambda x: x.name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def nick_name(self) -> str:
        return self._nick_name

    @property
    def desc(self) -> str:
        return self._desc

    @property
    def category(self) -> DeviceCategory:
        return self._category

    @property
    def middleware_name(self) -> str:
        return self._middleware_name

    @property
    def last_alive_time(self) -> float:
        return self._last_alive_time

    @property
    def alive_cycle(self) -> float:
        return self._alive_cycle

    @property
    def subscribed_topic_set(self) -> Set[str]:
        return self._subscribed_topic_set

    @property
    def is_super(self) -> bool:
        return self._is_super

    @property
    def is_parallel(self) -> bool:
        return self._is_parallel

    @property
    def is_builtin(self) -> bool:
        return self._is_builtin

    @property
    def is_manager(self) -> bool:
        return self._is_manager

    @property
    def is_matter(self) -> bool:
        return self._is_matter

    @property
    def request_client_id(self) -> str:
        return self._request_client_id

    @property
    def dependency_table(self) -> Dict[str, List[MXValue]]:
        return self._dependency_table

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @name.setter
    def name(self, name: str) -> None:
        self._name = name
        for service in self.service_list:
            service.thing_name = name

    @nick_name.setter
    def nick_name(self, nick_name: str) -> None:
        self._nick_name = nick_name

    @desc.setter
    def desc(self, desc: str) -> None:
        self._desc = desc

    @category.setter
    def category(self, category: DeviceCategory) -> None:
        self._category = category

    @middleware_name.setter
    def middleware_name(self, middleware_name: str) -> None:
        self._middleware_name = middleware_name
        for service in self.service_list:
            service.middleware_name = middleware_name

    @last_alive_time.setter
    def last_alive_time(self, last_alive_time: float) -> None:
        self._last_alive_time = last_alive_time

    @alive_cycle.setter
    def alive_cycle(self, alive_cycle: float) -> None:
        self._alive_cycle = alive_cycle

    @subscribed_topic_set.setter
    def subscribe_topic_set(self, subscribe_topic_set: Set[str]) -> None:
        self._subscribed_topic_set = subscribe_topic_set

    @is_super.setter
    def is_super(self, is_super: bool) -> bool:
        self._is_super = is_super

    @is_parallel.setter
    def is_parallel(self, is_parallel: bool) -> bool:
        self._is_parallel = is_parallel

    @is_builtin.setter
    def is_builtin(self, is_builtin: bool) -> bool:
        self._is_builtin = is_builtin

    @is_manager.setter
    def is_manager(self, is_manager: bool) -> bool:
        self._is_manager = is_manager

    @is_matter.setter
    def is_matter(self, is_matter: bool) -> bool:
        self._is_matter = is_matter

    @request_client_id.setter
    def request_client_id(self, request_client_id: str) -> None:
        self._request_client_id = request_client_id

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def generate_thing_id(self, name: str, append_mac_address: bool, interface: str = None) -> str:
        mac_address = get_mac_address(interface=interface)

        if not check_valid_identifier(name):
            raise MXValueError(f'name cannot be empty & can only contain alphanumeric characters and underscores. name: {self._name}')

        if not append_mac_address:
            return f'{name}'
        elif mac_address:
            return f'{name}_{mac_address}'
        else:
            mac = [random.randint(0x00, 0xFF) for _ in range(6)]
            rand_mac_address = ''.join(map(lambda x: '%02x' % x, mac)).upper()
            return f'{name}_{rand_mac_address}'

    def generate_register_message(self) -> MXRegisterMessage:
        if self.is_super:
            for function in self.function_list:
                if not hasattr(function, '_sub_service_request_list'):
                    return False
        else:
            for function in self.function_list:
                if hasattr(function, '_sub_service_request_list'):
                    return False

        for function in self.function_list:
            for arg in function.arg_list:
                bound = arg.bound
                if not arg._name:
                    return False
                if not arg.type in [MXType.STRING, MXType.BINARY] and bound[1] - bound[0] <= 0:
                    return False

        for value in self.value_list:
            bound = value.bound
            if not value.type in [MXType.STRING, MXType.BINARY] and bound[1] - bound[0] <= 0:
                return False

        register_msg = MXRegisterMessage(thing_name=self.name, payload=self.dict())
        return register_msg

    def generate_unregister_message(self) -> MXUnregisterMessage:
        return MXUnregisterMessage(self.name)

    def generate_alive_message(self) -> MXAliveMessage:
        return MXAliveMessage(self.name)

    # need for middleware detect
    def generate_get_home_message(self) -> MXGetHomeMessage:
        return MXGetHomeMessage(self.name)



from big_thing_py.utils.log_util import *

import traceback


def print_error(e: Exception):
    traceback_msg = traceback.format_exc()
    MXLOG_DEBUG(f'Traceback message : {traceback_msg}\nError: {e}', 'red')



from big_thing_py.common.common import *

from threading import Thread, Event, Lock, Timer
from queue import Queue, Empty


class ThreadMode(Enum):
    NORMAL = auto()
    TIMER = auto()


class MXThread:
    def __init__(
        self,
        name: str = None,
        target: Callable = None,
        args: Tuple = (),
        kwargs: dict = {},
        daemon: bool = True,
        mode: ThreadMode = ThreadMode.NORMAL,
        interval: float = None,
    ) -> None:
        self._name: str = name
        self._target: Callable = target
        self._args: Tuple = args
        self._kwargs: dict = kwargs
        self._daemon: List[str] = daemon
        self._mode: ThreadMode = mode
        self._thread: Thread = None

        if not target:
            raise Exception('[MXThread] No callback function to run')
        if self._mode == ThreadMode.TIMER and interval is None:
            raise Exception('[MXThread] Thread mode is Timer, but interval is not given...')

        self.set_thread()

    def set_thread(self) -> None:
        if isinstance(self._args, tuple):
            self._args: list = list(self._args)
        else:
            self._args = []

        self._args = tuple(self._args)

        if not self._name:
            self._name = '_'.join(self._target.__name__.split('_')[:-1])

        if self._mode == ThreadMode.NORMAL:
            self._thread = Thread(target=self._target, name=self._name, args=self._args, kwargs=self._kwargs, daemon=self._daemon)
        elif self._mode == ThreadMode.TIMER:
            self._thread = Timer(target=self._target, name=self._name, args=self._args, kwargs=self._kwargs, daemon=self._daemon)
        else:
            raise Exception('[MXThread] Invalid thread mode')

    def start(self) -> None:
        self._thread.start()

    def join(self) -> None:
        self._thread.join()

    def is_alive(self) -> bool:
        return self._thread.is_alive()

    def get_name(self) -> str:
        return self._name

    def get_target_function_name(self) -> str:
        return self._name.split('_')[0]

    def get_target_scenario_name(self) -> str:
        return self._name.split('_')[1]



from big_thing_py.core.thing import *
from big_thing_py.core.argument import MXArgument
from big_thing_py.core.mqtt_message import (
    MXExecuteResultMessage,
    MXRegisterResultMessage,
    MXRequestUnregisterMessage,
    MXBinaryValueResultMessage,
    MXRequestRegisterInfoMessage,
    MXExecuteMessage,
    MXUnregisterResultMessage,
    MXHomeResultMessage,
)
from big_thing_py.core.request import MXExecuteRequest, MXInnerExecuteRequest
from big_thing_py.core.ble_advertiser import BLEAdvertiser, BLEErrorCode
from big_thing_py.core.wifi_manager import WiFiManager
from big_thing_py.core.device_model import Objects as DeviceCategory

import ssl
import asyncio
import uvloop
import zeroconf
from zeroconf import ServiceInfo, Zeroconf
import socket

from gmqtt import Client as MQTTClient

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


# ASCII Art from https://patorjk.com/software/taag/#p=display&h=1&v=1&f=Big

# ======================================================================== #
#    _____         _____   ____   _      _______  _      _                 #
#   / ____|       |  __ \ |  _ \ (_)    |__   __|| |    (_)                #
#  | (___    ___  | |__) || |_) | _   __ _ | |   | |__   _  _ __    __ _   #
#   \___ \  / _ \ |  ___/ |  _ < | | / _` || |   | '_ \ | || '_ \  / _` |  #
#   ____) || (_) || |     | |_) || || (_| || |   | | | || || | | || (_| |  #
#  |_____/  \___/ |_|     |____/ |_| \__, ||_|   |_| |_||_||_| |_| \__, |  #
#                                     __/ |                         __/ |  #
#                                    |___/                         |___/   #
# ======================================================================== #


class ThingState(Enum):
    RESET = auto()
    NETWORK_SETUP = auto()
    BLE_ADVERTISE = auto()
    BROKER_SETUP = auto()
    NETWORK_CONNECTED = auto()
    BROKER_CONNECTED = auto()
    REGISTERED = auto()
    RUNNING = auto()
    UNREGISTERED = auto()
    RECONNECTED = auto()
    NETWORK_LOST = auto()
    BROKER_LOST = auto()
    NETWORK_RECONNECTED = auto()
    BROKER_RECONNECTED = auto()
    SHUTDOWN = auto()


class MXBigThing:
    MIDDLEWARE_ONLINE_CHECK_INTERVAL = 0.5
    MIDDLEWARE_RECONNECT_TIMEOUT = 30
    CONNECT_RETRY = 3
    ALIVE_CYCLE_SCALER = 2.1
    HANDLE_MQTT_MESSAGE_IGNORE_ERROR_LIST = [
        MXErrorCode.NO_ERROR,
        MXErrorCode.DUPLICATE,
        MXErrorCode.NOT_FOUND_ERROR,
    ]

    def __init__(
        self,
        name: str = MXThing.DEFAULT_NAME,
        nick_name: str = MXThing.DEFAULT_NAME,
        category: DeviceCategory = DeviceCategory.Undefined,
        desc='',
        version=sdk_version(),
        service_list: List[MXService] = [],
        alive_cycle: float = 60,
        is_super: bool = False,
        is_parallel: bool = True,
        is_builtin: bool = False,
        is_manager: bool = False,
        is_matter: bool = False,
        ip: str = '127.0.0.1',
        port: int = 1883,
        ssl_ca_path: str = '',
        ssl_cert_path: str = '',
        ssl_key_path: str = '',
        log_name: str = '',
        log_enable: bool = True,
        log_mode: MXPrintMode = MXPrintMode.ABBR,
        append_mac_address: bool = True,
    ):
        # Thing
        self._thing_data: MXThing = MXThing(
            name=name,
            nick_name=nick_name,
            category=category,
            desc=desc,
            version=version,
            service_list=service_list,
            alive_cycle=alive_cycle,
            is_super=is_super,
            is_parallel=is_parallel,
            is_builtin=is_builtin,
            is_manager=is_manager,
            is_matter=is_matter,
        )

        # Log
        self._log_mode = log_mode
        self._log_enable = log_enable
        self._log_name = log_name

        # WiFi
        self._ssid = ''
        self._pw = ''

        # MQTT
        self._mqtt_client: MQTTClient = None
        self._ip = ip
        self._port = port
        self._ssl_ca_path = ssl_ca_path
        self._ssl_cert_path = ssl_cert_path
        self._ssl_key_path = ssl_key_path

        # Util
        self._append_mac_address = append_mac_address

        # Queue
        # self._receive_queue: asyncio.Queue = asyncio.Queue()
        self._receive_queue: Dict[MXProtocolType, asyncio.Queue] = {
            MXProtocolType.Base.MT_REQUEST_REGISTER_INFO: asyncio.Queue(),
            MXProtocolType.Base.MT_REQUEST_UNREGISTER: asyncio.Queue(),
            MXProtocolType.Base.MT_RESULT_REGISTER: asyncio.Queue(),
            MXProtocolType.Base.MT_RESULT_UNREGISTER: asyncio.Queue(),
            MXProtocolType.Base.MT_RESULT_BINARY_VALUE: asyncio.Queue(),
            MXProtocolType.Base.MT_EXECUTE: asyncio.Queue(),
            MXProtocolType.Base.MT_IN_EXECUTE: asyncio.Queue(),
            MXProtocolType.WebClient.ME_RESULT_HOME: asyncio.Queue(),
        }

        # Tasks
        self._g_exit = asyncio.Event()
        self._task_list: List[asyncio.Task] = []

        # function_name: MXValue
        # if function executed, MXValue's inner value will be updated
        self._execute_task_list: List[asyncio.Task] = []

        self._loop = None
        self._state = ThingState.RESET
        self._ble_advertiser = BLEAdvertiser(server_name=f'Joi Smart Device {get_mac_address()}')
        self._wifi_manager = WiFiManager()
        self._zeroconf = Zeroconf()
        self._mqtt_mdns_service: ServiceInfo = None
        self._mqtt_ssl_mdns_service: ServiceInfo = None

        if self.is_super and not self.is_parallel:
            raise MXValueError('Super Thing must be parallel')

    def __eq__(self, o: 'MXBigThing') -> bool:
        instance_check = isinstance(o, MXBigThing)
        is_parallel_check = self.is_parallel == o.is_parallel
        is_super_check = self.is_super == o.is_super

        return super().__eq__(o) and instance_check and is_parallel_check and is_super_check

    def __getstate__(self):
        state = super().__getstate__()

        state['_ip'] = self._ip
        state['_port'] = self._port
        state['_ssl_ca_path'] = self._ssl_ca_path
        state['_ssl_cert_path'] = self._ssl_cert_path
        state['_ssl_key_path'] = self._ssl_key_path
        state['_append_mac_address'] = self._append_mac_address

        del state['_log_mode']
        del state['_log_enable']
        del state['_log_name']
        del state['_mqtt_client']
        del state['_g_exit']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._ip = state['_ip']
        self._port = state['_port']
        self._ssl_ca_path = state['_ssl_ca_path']
        self._ssl_cert_path = state['_ssl_cert_path']
        self._ssl_key_path = state['_ssl_key_path']
        self._append_mac_address = state['_append_mac_address']

        self._log_mode = MXPrintMode.ABBR
        self._log_enable = True
        self._log_name = ''
        self._mqtt_client = None
        self._g_exit = asyncio.Event()

    async def _setup(self) -> 'MXBigThing':
        self._ip = convert_url_to_ip(self._ip)
        if not is_valid_ip_address(self._ip) or not 1024 < self._port <= 65535:
            raise MXValueError('Invalid IP address, port number')
        else:
            self._ip = convert_url_to_ip(self._ip)

        if not check_python_version():
            raise MXNotSupportedError('Python version must be 3.7 or higher')

        self._init_logger()
        self._update_thing_id(self.name)

        for service in self.service_list:
            service.add_tag(MXTag(self.name))

        self._mqtt_mdns_service = ServiceInfo(
            type_='_mqtt._tcp.local.',
            name=f'_myssix_thing_{self.name}._mqtt._tcp.local.',
            addresses=[socket.inet_aton(get_local_ip())],
            port=1883,
            properties={'info': 'MySSIX Thing for Non-SSL', 'thing_id': self.name},
        )
        self._mqtt_ssl_mdns_service = ServiceInfo(
            type_='_mqtt-ssl._tcp.local.',
            name=f'_myssix_thing_{self.name}._mqtt-ssl._tcp.local.',
            addresses=[socket.inet_aton(get_local_ip())],
            port=8883,
            properties={'info': 'MySSIX Thing for SSL', 'thing_id': self.name},
        )

        self._mqtt_client = MQTTClient(client_id=self.name)
        # FIXME: Later, username and password options must be given by argument of the MXBigThing class
        self._mqtt_client.set_auth_credentials('tester', 'test12')
        self._mqtt_client.set_config({'reconnect_delay': 1})

        self._loop = asyncio.get_event_loop()

        return self

    async def run(self):
        await self._setup()
        await self._main_event_loop()

    async def _wrapup(self):
        try:
            if self._mqtt_client and self._mqtt_client.is_connected:
                self._send_TM_UNREGISTER(self._thing_data)
                # FIXME: Need to wait for the result of unregister
                # recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_UNREGISTER].get()
                # error = self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

        except Exception as e:
            print_error(e)
            return False
        finally:
            if await self._ble_advertiser.is_advertising():
                await self._ble_advertiser.stop()

            self._unregister_mdns_service()

            if self._mqtt_client and self._mqtt_client.is_connected:
                await self._disconnect_from_broker(disconnect_try=MXBigThing.CONNECT_RETRY)

            MXLOG_DEBUG('Thing Exit', 'red')
            sys.exit(0)

    # ================================================================================================= #
    #   _______  _                            _   ______                    _    _                      #
    #  |__   __|| |                          | | |  ____|                  | |  (_)                     #
    #     | |   | |__   _ __  ___   __ _   __| | | |__  _   _  _ __    ___ | |_  _   ___   _ __   ___   #
    #     | |   | '_ \ | '__|/ _ \ / _` | / _` | |  __|| | | || '_ \  / __|| __|| | / _ \ | '_ \ / __|  #
    #     | |   | | | || |  |  __/| (_| || (_| | | |   | |_| || | | || (__ | |_ | || (_) || | | |\__ \  #
    #     |_|   |_| |_||_|   \___| \__,_| \__,_| |_|    \__,_||_| |_| \___| \__||_| \___/ |_| |_||___/  #
    # ================================================================================================= #

    async def _preprocessing(self):
        await asyncio.sleep(THREAD_TIME_OUT)

        if self._execute_task_list:
            tasks = await asyncio.wait(self._execute_task_list, return_when=asyncio.FIRST_COMPLETED)
            done_tasks: List[asyncio.Task] = tasks[0]
            for task in done_tasks:
                exception = task.exception()
                if exception:
                    print_error(exception)
                else:
                    execute_request: MXExecuteRequest = task.result()

                    error = self._send_TM_RESULT_EXECUTE(execute_request, self._thing_data)
                    if error == MXErrorCode.NOT_FOUND_ERROR:
                        MXLOG_DEBUG(f'[{get_current_function_name()}] Target function not found!!!', 'red')
                        continue
                    elif error != MXErrorCode.NO_ERROR:
                        MXLOG_DEBUG(f'[{get_current_function_name()}] Send function result failed!!!', 'red')
                        continue

                    dependency_table = self._thing_data.dependency_table
                    dependency_info_list = dependency_table.get(execute_request.result_msg.function_name, [])
                    for dependency_info in dependency_info_list:
                        value: MXValue = dependency_info['value']
                        if dependency_info['set_value_as_result']:
                            target_function = self._get_function(execute_request.result_msg.function_name)
                            set_value: Union[int, float, bool, str] = target_function.return_value
                        else:
                            set_value: Union[int, float, bool, str] = dependency_info['set_value']

                        new_value = await value.update(set_value=set_value)
                        if new_value is None:
                            continue

                        self._send_TM_VALUE_PUBLISH(value)

                        set_value_with_cond = dependency_info['set_value_with_cond']
                        if set_value_with_cond:
                            set_value = set_value_with_cond(new_value)
                        else:
                            set_value: Union[int, float, bool, str] = dependency_info['set_value']
                            new_value = await value.update(set_value=set_value)
                            if new_value is None:
                                continue

            self._execute_task_list = [task for task in self._execute_task_list if not task.done()]

    async def _RESET_state_process(self):
        if self.is_builtin or self.is_manager:
            self.next_state = ThingState.BROKER_SETUP
        else:
            self.next_state = ThingState.BLE_ADVERTISE

    async def _BLE_ADVERTISE_state_process(self):
        # BLE Advertise
        await self._ble_advertiser.start(self.name)
        if not await self._ble_advertiser.is_advertising():
            MXLOG_DEBUG(f'BLE Advertiser start failed...', 'red')
            await self._ble_advertiser.stop()
            self.next_state = ThingState.RESET
            return

        # Save WiFi, Broker info
        MXLOG_DEBUG(f'Wait for WiFi credentials from BLE...', 'yellow')
        self._ssid, self._pw, broker, ble_error = await self._ble_advertiser.wait_until_wifi_credentials_set(timeout=None)
        self._ip = broker.split(':')[0]
        self._port = int(broker.split(':')[1])

        if ble_error != BLEErrorCode.NO_ERROR:
            MXLOG_DEBUG(f'Something getting wrong while BLE setup! error code: {ble_error}', 'red')
            await self._ble_advertiser.stop()
            self.next_state = ThingState.RESET
            return

        await self._ble_advertiser.stop()
        self.next_state = ThingState.NETWORK_SETUP

    async def _NETWORK_SETUP_state_process(self):
        # WiFi Connect
        self._wifi_manager.ssid = self._ssid
        self._wifi_manager.password = self._pw
        await self._wifi_manager.connect()
        if self._wifi_manager.check_connection() and self._wifi_manager.network_reachable(self._ip, self._port):
            MXLOG_DEBUG(f'WiFi connection success. SSID: {self._wifi_manager.get_connected_wifi_ssid()}', 'green')
            self.next_state = ThingState.BROKER_SETUP
        else:
            if wifi_connect_try > 0:
                MXLOG_DEBUG(f'Connect to SSID {self._wifi_manager.ssid} failed... (try: {wifi_connect_try})', 'yellow')
                wifi_connect_try -= 1
            else:
                MXLOG_DEBUG(f'WiFi connection failed... Go back to BLE setup.', 'red')
                wifi_connect_try = MXBigThing.CONNECT_RETRY
                self.next_state = ThingState.RESET

    async def _BROKER_SETUP_state_process(self):
        if not await self._connect_to_broker(MXBigThing.CONNECT_RETRY):
            MXLOG_DEBUG(f'Connect to broker failed... Check broker host and port -> {self._ip}:{self._port}', 'red')
            self.next_state = ThingState.SHUTDOWN
            return

        # Setup avahi mDNS service
        if not self._register_mdns_service():
            MXLOG_DEBUG(f'Register mDNS service failed... Retry to register mDNS service.', 'red')
            self._unregister_mdns_service()
            self.next_state = ThingState.BROKER_SETUP
            return

        self.next_state = ThingState.BROKER_CONNECTED

    async def _BROKER_CONNECTED_state_process(self):
        # Prepare to register
        self._subscribe_init_topic_list(self._thing_data)
        if self.is_builtin or self.is_manager:
            MXLOG_DEBUG(f'Run builtin or plugin Thing', 'yellow')
            self._send_TM_REGISTER(self._thing_data)

            recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_REGISTER].get()
            self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)
        else:
            recv_msg = await self._receive_queue[MXProtocolType.Base.MT_REQUEST_REGISTER_INFO].get()
            self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

            recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_REGISTER].get()
            self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

    async def _REGISTERED_state_process(self):
        self.next_state = ThingState.RUNNING

    async def _RUNNING_state_process(self):
        # MQTT receive handling
        if not self._receive_queue_empty():
            recv_msg = await self._receive_queue_get()
            error = self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)
            if not error in MXBigThing.HANDLE_MQTT_MESSAGE_IGNORE_ERROR_LIST:
                MXLOG_DEBUG(f'[{get_current_function_name()}] MQTT Message handling failed', 'red')

        # Value publish handling
        current_time = get_current_datetime()
        for value in self.value_list:
            if not value.is_initialized:
                await value.update()
                self._send_TM_VALUE_PUBLISH(value)
            # NOTE (thsvkd): If Thing is not Manager of Staff, I think event-based value feature is not needed...
            # elif current_time - value.last_update_time > value.cycle and value.cycle != 0:
            elif current_time - value.last_update_time > value.cycle:
                new_value = await value.update()
                if new_value is None:
                    continue

                self._send_TM_VALUE_PUBLISH(value)

        # Alive handling
        if current_time - self.last_alive_time > self.alive_cycle / MXBigThing.ALIVE_CYCLE_SCALER:
            self._send_TM_ALIVE(self._thing_data)

    async def _UNREGISTERED_state_process(self):
        if await self._ble_advertiser.is_advertising():
            await self._ble_advertiser.stop()

        self._unregister_mdns_service()

        if self._mqtt_client and self._mqtt_client.is_connected:
            await self._disconnect_from_broker(disconnect_try=MXBigThing.CONNECT_RETRY)

        # TODO: need to disconnect from wifi router

        self.next_state = ThingState.RESET

    async def _RECONNECTED_state_process(self):
        # TODO: do something
        pass

    async def _NETWORK_LOST_state_process(self):
        # WiFi Connect
        self._wifi_manager.ssid = self._ssid
        self._wifi_manager.password = self._pw
        await self._wifi_manager.connect()
        if self._wifi_manager.check_connection():
            MXLOG_DEBUG(f'WiFi connection success. SSID: {self._wifi_manager.get_connected_wifi_ssid()}', 'green')
            self.next_state = ThingState.BROKER_LOST
        else:
            if wifi_connect_try > 0:
                MXLOG_DEBUG(f'Connect to SSID {self._wifi_manager.ssid} failed... (try: {wifi_connect_try})', 'yellow')
                wifi_connect_try -= 1
                self.next_state = ThingState.NETWORK_LOST
            else:
                MXLOG_DEBUG(f'WiFi connection failed... Go back to BLE setup.', 'red')
                wifi_connect_try = MXBigThing.CONNECT_RETRY
                self.next_state = ThingState.RESET

        self.next_state = ThingState.BROKER_LOST

    async def _BROKER_LOST_state_process(self):
        if self._mqtt_client.is_connected:
            self.next_state = ThingState.BROKER_RECONNECTED
        else:
            self.next_state = ThingState.BROKER_LOST
            await asyncio.sleep(1)

    async def _NETWORK_RECONNECTED_state_process(self):
        # TODO: do something
        pass

    async def _BROKER_RECONNECTED_state_process(self):
        get_home_msg = self._thing_data.generate_get_home_message().mqtt_message()
        result_home_topic = MXProtocolType.WebClient.ME_RESULT_HOME.value % '+'
        self._subscribe(result_home_topic)

        current_time = get_current_datetime()
        if get_current_datetime() - current_time < MXBigThing.MIDDLEWARE_RECONNECT_TIMEOUT:
            self._publish(get_home_msg.topic, get_home_msg.payload)
            try:
                recv_msg = self._receive_queue[MXProtocolType.WebClient.ME_RESULT_HOME].get_nowait()
                self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

                # Auto subscription restore feature
                for topic in self.subscribed_topic_set:
                    self._subscribe(topic)

                self._send_TM_ALIVE(self._thing_data)
            except asyncio.QueueEmpty:
                time.sleep(MXBigThing.MIDDLEWARE_ONLINE_CHECK_INTERVAL)
        else:
            if self.is_builtin or self.is_manager:
                self.next_state = ThingState.SHUTDOWN
            else:
                MXLOG_DEBUG(f'Middleware is offline... Go back to BLE setup.', 'red')
                self.next_state = ThingState.BLE_ADVERTISE

    async def _main_event_loop(self):
        while not self._g_exit.is_set():
            try:
                await self._preprocessing()

                if self._state == ThingState.RESET:
                    await self._RESET_state_process()
                elif self._state == ThingState.BLE_ADVERTISE:
                    await self._BLE_ADVERTISE_state_process()
                elif self._state == ThingState.NETWORK_SETUP:
                    await self._NETWORK_SETUP_state_process()
                elif self._state == ThingState.BROKER_SETUP:
                    await self._BROKER_SETUP_state_process()
                elif self._state == ThingState.BROKER_CONNECTED:
                    await self._BROKER_CONNECTED_state_process()
                elif self._state == ThingState.REGISTERED:
                    await self._REGISTERED_state_process()
                elif self._state == ThingState.RUNNING:
                    await self._RUNNING_state_process()
                elif self._state == ThingState.UNREGISTERED:
                    await self._UNREGISTERED_state_process()
                elif self._state == ThingState.RECONNECTED:
                    await self._RECONNECTED_state_process()
                elif self._state == ThingState.NETWORK_LOST:
                    await self._NETWORK_LOST_state_process()
                elif self._state == ThingState.BROKER_LOST:
                    await self._BROKER_LOST_state_process()
                elif self._state == ThingState.NETWORK_RECONNECTED:
                    await self._NETWORK_RECONNECTED_state_process()
                elif self._state == ThingState.BROKER_RECONNECTED:
                    await self._BROKER_RECONNECTED_state_process()
                elif self._state == ThingState.SHUTDOWN:
                    await self._wrapup()
                else:
                    MXLOG_DEBUG(f'Unexpected state!!! state: {self._state}', 'red')
                    self.next_state = ThingState.SHUTDOWN
            except BaseException as e:
                if isinstance(e, KeyboardInterrupt):
                    MXLOG_DEBUG('KeyboardInterrupt Exit', 'red')
                    await self._wrapup()
                elif isinstance(e, asyncio.CancelledError):
                    MXLOG_DEBUG('asyncio.CancelledError Exit', 'red')
                    await self._wrapup()
                elif isinstance(e, zeroconf._exceptions.NonUniqueNameException):
                    self._unregister_mdns_service()
                    self.next_state = ThingState.BROKER_SETUP
                elif isinstance(e, ConnectionRefusedError):
                    MXLOG_DEBUG(f'Connect to broker failed... broker - {self._ip}:{self._port}', 'red')
                elif isinstance(e, ConnectionResetError):
                    MXLOG_DEBUG(f'Disconnect from broker failed... broker - {self._ip}:{self._port}', 'red')
                else:
                    print_error(e)
                    raise e

    # ======================================================================================================================= #
    #  _    _                    _  _         __  __   ____  _______  _______   __  __                                        #
    # | |  | |                  | || |       |  \/  | / __ \|__   __||__   __| |  \/  |                                       #
    # | |__| |  __ _  _ __    __| || |  ___  | \  / || |  | |  | |      | |    | \  / |  ___  ___  ___   __ _   __ _   ___    #
    # |  __  | / _` || '_ \  / _` || | / _ \ | |\/| || |  | |  | |      | |    | |\/| | / _ \/ __|/ __| / _` | / _` | / _ \   #
    # | |  | || (_| || | | || (_| || ||  __/ | |  | || |__| |  | |      | |    | |  | ||  __/\__ \\__ \| (_| || (_| ||  __/   #
    # |_|  |_| \__,_||_| |_| \__,_||_| \___| |_|  |_| \___\_\  |_|      |_|    |_|  |_| \___||___/|___/ \__,_| \__, | \___|   #
    #                                                                                                         __/ |           #
    #                                                                                                         |___/           #
    # ======================================================================================================================= #

    def _handle_mqtt_message(self, msg: MQTTMessage, target_thing: MXThing, state_change: bool = True) -> MXErrorCode:
        topic = decode_MQTT_message(msg)[0]
        protocol = MXProtocolType.get(topic)

        if protocol == MXProtocolType.Base.MT_REQUEST_REGISTER_INFO:
            error = self._handle_MT_REQUEST_REGISTER_INFO(msg, target_thing=target_thing)
        elif protocol == MXProtocolType.Base.MT_REQUEST_UNREGISTER:
            error = self._handle_MT_REQUEST_UNREGISTER(msg, target_thing=target_thing)
        elif protocol == MXProtocolType.Base.MT_RESULT_REGISTER:
            error = self._handle_MT_RESULT_REGISTER(msg, target_thing=target_thing)
            if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
                # NOTE (thsvkd): MT_RESULT_REGISTER topic also use as a trigger for sending the current
                # Value to the middleware regardless of the cycle. If the MT_RESULT_REGISTER topic comes
                # while running after being registered, it is implemented to work as a trigger for value
                # publish by initializing the values.
                if self._state in [ThingState.RUNNING, ThingState.REGISTERED]:
                    for value in target_thing.value_list:
                        value.is_initialized = False

                self.next_state = ThingState.REGISTERED if state_change else self._state
            elif error == MXErrorCode.INVALID_DESTINATION:
                self.next_state = ThingState.REGISTERED if state_change else self._state
            elif error == MXErrorCode.FAIL:
                MXLOG_DEBUG(
                    f'{PrintTag.ERROR} Register failed... Thing {self.name} register packet is not valid. error code: {error}',
                    'red',
                )
                self.next_state = ThingState.SHUTDOWN if state_change else self._state
            elif error == MXErrorCode.INVALID_REQUEST:
                MXLOG_DEBUG(f'Receive unexpected packet... Drop it. topic: {topic}', 'magenta')
            else:
                MXLOG_DEBUG(
                    f'{PrintTag.ERROR} Register failed... Thing {self.name} register packet is not valid. error code: {error}',
                    'red',
                )
                self.next_state = ThingState.SHUTDOWN if state_change else self._state
        elif protocol == MXProtocolType.Base.MT_RESULT_UNREGISTER:
            error = self._handle_MT_RESULT_UNREGISTER(msg, target_thing=target_thing)
            if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
                self.next_state = ThingState.UNREGISTERED if state_change else self._state
            elif error == MXErrorCode.INVALID_DESTINATION:
                self.next_state = ThingState.UNREGISTERED if state_change else self._state
            elif error == MXErrorCode.FAIL:
                MXLOG_DEBUG(
                    f'{PrintTag.ERROR} Unregister failed...',
                    'red',
                )
            elif error == MXErrorCode.INVALID_REQUEST:
                MXLOG_DEBUG(f'Receive unexpected packet... Drop it. topic: {topic}', 'magenta')
            else:
                MXLOG_DEBUG(
                    f'{PrintTag.ERROR} Unregister failed... Thing {self.name} Unregister packet is not valid. error code: {error}',
                    'red',
                )
                self.next_state = ThingState.SHUTDOWN if state_change else self._state
        elif protocol == MXProtocolType.Base.MT_RESULT_BINARY_VALUE:
            error = self._handle_MT_RESULT_BINARY_VALUE(msg, target_thing=target_thing)
        elif protocol in [MXProtocolType.Base.MT_EXECUTE, MXProtocolType.Base.MT_IN_EXECUTE]:
            error = self._handle_MT_EXECUTE(msg, target_thing=target_thing)
        # for Auto Re-register feature
        elif protocol == MXProtocolType.WebClient.ME_RESULT_HOME:
            error = self._handle_ME_HOME_RESULT(msg)

            result_home_topic = MXProtocolType.WebClient.ME_RESULT_HOME.value % '+'
            if error == MXErrorCode.NO_ERROR:
                self._unsubscribe(result_home_topic)
                self.next_state = ThingState.REGISTERED if state_change else self._state
        else:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Unexpected topic! topic: {topic}')
            error = MXErrorCode.FAIL

        return error

    # ===============
    # ___  ___ _____
    # |  \/  ||_   _|
    # | .  . |  | |
    # | |\/| |  | |
    # | |  | |  | |
    # \_|  |_/  \_/
    # ===============

    def _handle_MT_REQUEST_REGISTER_INFO(self, msg: MQTTMessage, target_thing: MXThing) -> MXErrorCode:
        request_register_info_msg = MXRequestRegisterInfoMessage(msg)

        if request_register_info_msg.topic_error or request_register_info_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        middleware_name = request_register_info_msg.middleware_name
        MXLOG_DEBUG(f'Receive register request from Middleware {middleware_name}! Send Register packet.', 'green')

        target_thing.request_client_id = request_register_info_msg.client_id
        self._send_TM_REGISTER(target_thing)
        return MXErrorCode.NO_ERROR

    def _handle_MT_REQUEST_UNREGISTER(self, msg: MQTTMessage, target_thing: MXThing) -> MXErrorCode:
        request_unregister_msg = MXRequestUnregisterMessage(msg)

        if request_unregister_msg.topic_error or request_unregister_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        middleware_name = request_unregister_msg.middleware_name
        MXLOG_DEBUG(f'Receive unregister request from Middleware {middleware_name}! Send Unregister packet.', 'green')

        target_thing.request_client_id = request_unregister_msg.client_id
        self._send_TM_UNREGISTER(target_thing)
        return MXErrorCode.NO_ERROR

    def _handle_MT_RESULT_REGISTER(self, msg: MQTTMessage, target_thing: MXThing) -> MXErrorCode:
        register_result_msg = MXRegisterResultMessage(msg)
        error = register_result_msg.error

        if register_result_msg.topic_error or register_result_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        if target_thing.name != register_result_msg.thing_name:
            # MXLOG_DEBUG(
            #     f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.name} should be arrive, not {register_result_msg.thing_name}'
            # )
            return MXErrorCode.INVALID_DESTINATION

        if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
            MXLOG_DEBUG(
                f'{PrintTag.GOOD if error == MXErrorCode.NO_ERROR else PrintTag.DUP} Thing {target_thing.name} {colored("register", "green")} success!'
            )
            target_thing.middleware_name = register_result_msg.middleware_name
            for service in target_thing.service_list:
                service.middleware_name = target_thing.middleware_name
                service.thing_name = target_thing.name
            self._subscribe_service_topic_list(target_thing)
        elif error == MXErrorCode.FAIL:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Register failed... Thing {target_thing.name} register packet is not valid. error code: {register_result_msg.error}',
                'red',
            )
        else:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Register failed... Unexpected error occurred!!! error code: {register_result_msg.error}',
                'red',
            )

        return error

    def _handle_MT_RESULT_UNREGISTER(self, msg: MQTTMessage, target_thing: MXThing) -> MXErrorCode:
        unregister_result_msg = MXUnregisterResultMessage(msg)
        error = unregister_result_msg.error

        if unregister_result_msg.topic_error or unregister_result_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        if target_thing.name != unregister_result_msg.thing_name:
            # MXLOG_DEBUG(
            #     f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.name} should be arrive, not {unregister_result_msg.thing_name}',
            #     'red',
            # )
            return MXErrorCode.INVALID_DESTINATION

        if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
            MXLOG_DEBUG(
                f'{PrintTag.GOOD if error == MXErrorCode.NO_ERROR else PrintTag.DUP} Thing {target_thing.name} {colored("unregister", "red")} success!'
            )
            self._unsubscribe_all_topic_list(target_thing)
        elif error == MXErrorCode.FAIL:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Unregister failed... Thing {target_thing.name} unregister packet is not valid. error code: {unregister_result_msg.error}',
                'red',
            )
        else:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Unregister failed... Unexpected error occurred!!! error code: {unregister_result_msg.error}',
                'red',
            )

        return error

    def _handle_MT_EXECUTE(self, msg: MQTTMessage, target_thing: MXThing) -> Tuple[asyncio.Task, MXErrorCode]:
        execute_msg = MXExecuteMessage(msg)
        if execute_msg.topic_error or execute_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        target_function = target_thing.get_function(execute_msg.function_name)
        if not target_function:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Target function not found! - topic: {decode_MQTT_message(msg)[0]}')
            return MXErrorCode.NOT_FOUND_ERROR
        if target_thing.name != execute_msg.thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.name} should be arrive, not {execute_msg.thing_name}',
                'red',
            )
            return MXErrorCode.INVALID_REQUEST
        if execute_msg.topic_error or execute_msg.payload_error:
            MXLOG_DEBUG(f'[{get_current_function_name()}] execute_msg error! - topic: {decode_MQTT_message(msg)[0]}{execute_msg.topic_error}')
            return MXErrorCode.INVALID_REQUEST

        execute_protocol = MXProtocolType.get(execute_msg.topic)
        execute_result_msg_input = dict(
            function_name=target_function.name,
            thing_name=target_function.thing_name,
            middleware_name=target_function.middleware_name,
            scenario=execute_msg.scenario,
            client_id=execute_msg.client_id,
            request_ID=execute_msg.request_ID,
            return_type=target_function.return_type,
            return_value=None,
        )
        if execute_protocol == MXProtocolType.Base.MT_EXECUTE:
            execute_request = MXExecuteRequest(
                trigger_msg=execute_msg,
                result_msg=MXExecuteResultMessage(
                    **execute_result_msg_input,
                    action_type=MXActionType.EXECUTE,
                ),
            )
        elif execute_protocol == MXProtocolType.Base.MT_IN_EXECUTE:
            execute_request = MXInnerExecuteRequest(
                trigger_msg=execute_msg,
                result_msg=MXExecuteResultMessage(
                    **execute_result_msg_input,
                    action_type=MXActionType.INNER_EXECUTE,
                ),
            )
        else:
            raise MXValueError('Invalid protocol type')

        # 서로의 arg_list가 일치하는 지 확인한다.
        if not self._compare_arg_list(target_function.arg_list, list(execute_request.trigger_msg.tuple_arguments())):
            self._send_TM_RESULT_EXECUTE(execute_request, self._thing_data)
            return MXErrorCode.EXECUTE_ARG_FAIL

        # 중복된 시나리오로부터 온 실행 요청이면 -4 에러코드를 보낸다.
        if execute_request.trigger_msg.scenario in target_function.running_scenario_list:
            execute_request.result_msg.error = MXErrorCode.DUPLICATE
            self._send_TM_RESULT_EXECUTE(execute_request, self._thing_data)
            return MXErrorCode.DUPLICATE

        # 병렬실행이 가능한지 옵션을 살펴보고, 이후에 실행 중인지 살펴본다.
        if target_thing.is_parallel:
            task = asyncio.create_task(target_function.execute(execute_request))
        elif not target_function._running:
            task = asyncio.create_task(target_function.execute(execute_request))
        else:
            execute_request.result_msg.error = MXErrorCode.FAIL
            self._send_TM_RESULT_EXECUTE(execute_request, self._thing_data)

        self._execute_task_list.append(task)

        return MXErrorCode.NO_ERROR

    def _handle_MT_RESULT_BINARY_VALUE(self, msg: MQTTMessage, target_thing: MXThing) -> bool:
        binary_value_result_msg = MXBinaryValueResultMessage(msg)

        if binary_value_result_msg.topic_error or binary_value_result_msg.payload_error:
            return MXErrorCode.INVALID_REQUEST

        if target_thing.name != binary_value_result_msg.thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.name} should be arrive, not {binary_value_result_msg.thing_name}',
                'red',
            )
            return MXErrorCode.INVALID_REQUEST

        for value in target_thing.value_list:
            if value.name == binary_value_result_msg.value_name and value.type == MXType.BINARY:
                value.binary_sending = False
                return MXErrorCode.NO_ERROR
        else:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Value {binary_value_result_msg.value_name} does not exist!!!', 'red')
            return MXErrorCode.NOT_FOUND_ERROR

    # ===============
    #  _____ ___  ___
    # |_   _||  \/  |
    #   | |  | .  . |
    #   | |  | |\/| |
    #   | |  | |  | |
    #   \_/  \_|  |_/
    # ===============

    def _send_TM_REGISTER(self, thing_data: MXThing) -> None:
        self._subscribe_init_topic_list(thing_data)

        register_msg = thing_data.generate_register_message()
        if not register_msg:
            raise Exception('TM_REGISTER packet error')

        register_mqtt_msg = register_msg.mqtt_message()
        self._publish(register_mqtt_msg.topic, register_mqtt_msg.payload)

    def _send_TM_UNREGISTER(self, thing_data: MXThing):
        unregister_msg = thing_data.generate_unregister_message()
        if not unregister_msg:
            raise Exception('TM_UNREGISTER packet error')

        unregister_mqtt_msg = unregister_msg.mqtt_message()
        self._publish(unregister_mqtt_msg.topic, unregister_mqtt_msg.payload)

    def _send_TM_ALIVE(self, thing_data: MXThing):
        alive_msg = thing_data.generate_alive_message()
        alive_mqtt_msg = alive_msg.mqtt_message()
        self._publish(alive_mqtt_msg.topic, alive_mqtt_msg.payload)
        thing_data.last_alive_time = get_current_datetime()

    def _send_TM_VALUE_PUBLISH(self, value: MXValue) -> None:
        value_publish_msg = value.generate_value_publish_message()
        value_publish_mqtt_msg = value_publish_msg.mqtt_message()

        if value.type == MXType.BINARY:
            value.binary_sending = True
        self._publish(value_publish_mqtt_msg.topic, value_publish_mqtt_msg.payload)

    def _send_TM_RESULT_EXECUTE(self, execute_request: Union[MXExecuteRequest, MXInnerExecuteRequest], thing_data: MXThing) -> MXErrorCode:
        target_function = thing_data.get_function(execute_request.result_msg.function_name)
        if target_function is None:
            return MXErrorCode.NOT_FOUND_ERROR

        execute_result_msg = target_function.generate_execute_result_message(execute_request)
        execute_result_mqtt_msg = execute_result_msg.mqtt_message()
        self._publish(execute_result_mqtt_msg.topic, execute_result_mqtt_msg.payload)
        return MXErrorCode.NO_ERROR

    # ================
    #  __  __  ______
    # |  \/  ||  ____|
    # | \  / || |__
    # | |\/| ||  __|
    # | |  | || |____
    # |_|  |_||______|
    # ================

    # for Auto Re-register feature
    def _handle_ME_HOME_RESULT(self, msg: MQTTMessage) -> MXErrorCode:
        execute_msg = MXHomeResultMessage(msg)
        if 'location' in execute_msg.payload:
            MXLOG_DEBUG(f'Middleware detected!!!')
            return MXErrorCode.NO_ERROR
        else:
            return MXErrorCode.FAIL

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def _get_function(self, function_name: str) -> MXFunction:
        return self._thing_data.get_function(function_name)

    def _receive_queue_empty(self):
        for queue in self._receive_queue.values():
            if not queue.empty():
                return False
        return True

    def _receive_queue_get(self):
        for queue in self._receive_queue.values():
            if not queue.empty():
                return queue.get()
        return None

    def _subscribe_init_topic_list(self, thing_data: MXThing) -> None:
        topic_list = [
            MXProtocolType.Base.MT_REQUEST_REGISTER_INFO.value % thing_data.name,
            MXProtocolType.Base.MT_REQUEST_UNREGISTER.value % thing_data.name,
            MXProtocolType.Base.MT_RESULT_REGISTER.value % thing_data.name,
            MXProtocolType.Base.MT_RESULT_UNREGISTER.value % thing_data.name,
            MXProtocolType.Base.MT_RESULT_BINARY_VALUE.value % thing_data.name,
        ]

        for topic in topic_list:
            self._subscribe(topic)

    def _subscribe_service_topic_list(self, thing_data: MXThing):
        topic_list = []

        for function in thing_data.function_list:
            topic_list += [
                MXProtocolType.Base.MT_EXECUTE.value % (function.name, thing_data.name, '+', '+'),
                (MXProtocolType.Base.MT_EXECUTE.value % (function.name, thing_data.name, '', '')).rstrip('/'),
                MXProtocolType.Base.MT_IN_EXECUTE.value % (function.name, thing_data.name),
            ]

        for topic in topic_list:
            self._subscribe(topic)

    def _unsubscribe_all_topic_list(self, thing_data: MXThing):
        # whenever _unsubscribe function execute, it remove target topic from self._subscribed_topic_set
        # so it need to iterate with copy of self._subscribed_topic_set
        target_topic_list = list(thing_data.subscribed_topic_set)
        for topic in target_topic_list:
            self._unsubscribe(topic)

    def _convert_arg_list(self, arg_list: List[Union[MXArgument, Union[int, float, str, bool]]]) -> List[MXType]:
        if isinstance(arg_list, list) and all([isinstance(arg, (int, float, str, bool)) for arg in arg_list]):
            arg_type_list = []
            for arg in arg_list:
                # if is_base64(arg):
                #     arg_type_list.append(MXType.BINARY)
                # else:
                #     arg_type_list.append(MXType.get(type(arg)))
                arg_type_list.append(MXType.get(type(arg)))
        elif isinstance(arg_list, list) and all([isinstance(arg, MXArgument) for arg in arg_list]):
            arg_type_list = [arg.type for arg in arg_list]
        else:
            raise MXTypeError(f'arg_list must be list of MXArgument or list of int, float, str, bool: {arg_list}')

        return arg_type_list

    def _compare_arg_list(
        self,
        arg_list1: List[Union[MXArgument, Union[int, float, str, bool]]],
        arg_list2: List[Union[MXArgument, Union[int, float, str, bool]]],
    ):
        arg_type_list1 = self._convert_arg_list(arg_list1)
        arg_type_list2 = self._convert_arg_list(arg_list2)

        if len(arg_type_list1) != len(arg_type_list2):
            MXLOG_DEBUG(f'Not matched arg_list length: {len(arg_type_list1)} != {len(arg_type_list2)}', 'red')
            return False

        for arg_type1, arg_type2 in zip(arg_type_list1, arg_type_list2):
            if arg_type1 in [MXType.INTEGER, MXType.DOUBLE] and arg_type2 in [MXType.INTEGER, MXType.DOUBLE]:
                pass
            elif arg_type1 != arg_type2:
                MXLOG_DEBUG(f'Not matched arg_list type: {arg_type1} != {arg_type2}', 'red')
                return False
        else:
            return True

    def _print_packet(
        self, topic: Union[str, bytes], payload: Union[str, bytes], direction: Direction, mode: MXPrintMode = MXPrintMode.FULL, pretty: bool = False
    ) -> str:
        if isinstance(topic, bytes):
            topic = topic.decode()
        if isinstance(payload, bytes):
            payload = dict_to_json_string(json_string_to_dict(payload.decode()), ensure_ascii=False)

        topic_template = MXProtocolType.get(topic)
        if not topic_template:
            MXLOG_DEBUG(f'[print_packet] Unknown topic!!! topic : {topic}')

        topic_indicator = '_'.join([topic_token for topic_token in topic_template.value.split('/') if topic_token != '%s'])
        payload = prune_payload(payload=dict_to_json_string(dict_object=payload, pretty=pretty), mode=mode)

        MXLOG_DEBUG(f'[{topic_indicator:20}][{direction.value}] topic: {topic} payload: {payload}')

    def add_service(self, service: MXService) -> 'MXThing':
        return self._thing_data.add_service(service)

    def dict(self) -> dict:
        return self._thing_data.dict()

    def add_dependency(
        self,
        function: MXFunction,
        value: MXValue,
        set_value: Union[int, float, bool, str] = None,
        set_value_as_result: bool = False,
    ) -> None:
        self._thing_data.add_dependency(function=function, value=value, set_value=set_value, set_value_as_result=set_value_as_result)

    #### MQTT utils ####

    async def _connect_to_broker(self, connect_try: int) -> bool:
        # Set SSL options
        if self._check_ssl_enable():
            MXLOG_DEBUG('Connect with SSL...', 'green')
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            ssl_context.load_cert_chain(certfile=self._ssl_cert_path, keyfile=self._ssl_key_path)
            ssl_context.load_verify_locations(cafile=self._ssl_ca_path)
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            # ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1  # TLSv1.2 이상만 사용
            ssl_context.check_hostname = False
        else:
            ssl_context = False
            MXLOG_DEBUG('[WARN] Connect without SSL...', 'yellow')

        # Broker Connect
        while connect_try:
            try:
                if self._mqtt_client.is_connected or await self._connect(self._ip, self._port, ssl=ssl_context):
                    return True

                await asyncio.sleep(THREAD_TIME_OUT * 10)
                MXLOG_DEBUG(f'Connect to Broker {self._ip}:{self._port} failed... (try: {connect_try})', 'yellow')
                connect_try -= 1
            except BaseException as e:
                if isinstance(e, ConnectionRefusedError) or isinstance(e, ConnectionResetError):
                    await asyncio.sleep(THREAD_TIME_OUT * 10)
                    MXLOG_DEBUG(f'Connect to Broker {self._ip}:{self._port} failed... (try: {connect_try})', 'yellow')
                    connect_try -= 1
                else:
                    raise e
        else:
            MXLOG_DEBUG(f'Broker connection failed...', 'red')
            connect_try = MXBigThing.CONNECT_RETRY
            return False

    async def _disconnect_from_broker(self, disconnect_try: int) -> bool:
        # Broker Disconnect
        while disconnect_try:
            try:
                await self._disconnect()
                if not self._mqtt_client.is_connected:
                    return True

                await asyncio.sleep(THREAD_TIME_OUT * 10)
                MXLOG_DEBUG(f'Disconnect to Broker {self._ip}:{self._port} failed... (try: {disconnect_try})', 'yellow')
                disconnect_try -= 1
            except BaseException as e:
                if isinstance(e, ConnectionResetError):
                    await asyncio.sleep(THREAD_TIME_OUT * 10)
                    MXLOG_DEBUG(f'Disconnect to Broker {self._ip}:{self._port} failed... (try: {disconnect_try})', 'yellow')
                    disconnect_try -= 1
                else:
                    raise e
        else:
            MXLOG_DEBUG(f'Broker disconnection failed... Go back to BLE setup.', 'red')
            disconnect_try = MXBigThing.CONNECT_RETRY
            return False

    async def _connect(self, ip: str, port: int, ssl: Union[ssl.SSLContext, bool], timeout: float = 1) -> bool:
        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_disconnect = self._on_disconnect
        # self._mqtt_client.on_publish = self._on_publish
        self._mqtt_client.on_subscribe = self._on_subscribe
        self._mqtt_client.on_unsubscribe = self._on_unsubscribe
        self._mqtt_client.on_message = self._on_message

        try:
            await asyncio.wait_for(self._mqtt_client.connect(ip, port, ssl=ssl), timeout=timeout)
        except asyncio.TimeoutError:
            return False

        return self._mqtt_client.is_connected

    async def _disconnect(self):
        await self._mqtt_client.disconnect()

    def _subscribe(self, topic: str, qos: int = 0):
        if not topic in self.subscribed_topic_set:
            rc = self._mqtt_client.subscribe(topic, qos=qos)
            self.subscribed_topic_set.add(topic)
            MXLOG_DEBUG(f'{PrintTag.SUBSCRIBE} {topic}')

        return True

    def _unsubscribe(self, topic: str):
        if topic in self.subscribed_topic_set:
            rc = self._mqtt_client.unsubscribe(topic)
            self.subscribed_topic_set.remove(topic)
            MXLOG_DEBUG(f'{PrintTag.UNSUBSCRIBE} {topic}')

        return True

    def _publish(self, topic: Union[str, bytes], payload: Union[str, bytes], qos: int = 0) -> int:
        self._print_packet(topic=topic, payload=payload, direction=Direction.PUBLISH, mode=self._log_mode)

        if isinstance(topic, str):
            topic = topic.encode()
        if isinstance(payload, str):
            payload = payload.encode()

        self._mqtt_client.publish(topic, payload, qos=qos)

    def _init_logger(self):
        if self._log_enable:
            START_LOGGER(whole_log_path=self._log_name, logging_mode=MXLogger.LoggingMode.ALL)
        else:
            START_LOGGER(logging_mode=MXLogger.LoggingMode.CONSOLE)

    def _check_ssl_enable(self) -> bool:
        if not self._ssl_ca_path and not self._ssl_cert_path and not self._ssl_key_path:
            return False
        if not os.path.exists(self._ssl_ca_path):
            raise MXNotFoundError(f'SSL CA file not found. invalid path: {self._ssl_ca_path}')
        elif not os.path.exists(self._ssl_cert_path):
            raise MXNotFoundError(f'SSL Cert file not found. invalid path: {self._ssl_cert_path}')
        elif not os.path.exists(self._ssl_key_path):
            raise MXNotFoundError(f'SSL Key file not found.. invalid path: {self._ssl_key_path}')
        else:
            return True

    def _update_thing_id(self, name: str, interface: str = None) -> None:
        self.name = self._thing_data.generate_thing_id(name, self._append_mac_address, interface=interface)

    def _register_mdns_service(self) -> bool:
        try:
            self._zeroconf.register_service(self._mqtt_mdns_service)
            self._zeroconf.register_service(self._mqtt_ssl_mdns_service)
            return True
        except zeroconf.NonUniqueNameException as e:
            return True
        except Exception as e:
            print_error(e)
            return False

    def _unregister_mdns_service(self) -> None:
        self._zeroconf.unregister_service(self._mqtt_mdns_service)
        self._zeroconf.unregister_service(self._mqtt_ssl_mdns_service)

    # ===================================================================================
    # ___  ___ _____  _____  _____   _____         _  _  _                   _
    # |  \/  ||  _  ||_   _||_   _| /  __ \       | || || |                 | |
    # | .  . || | | |  | |    | |   | /  \/  __ _ | || || |__    __ _   ___ | | __ ___
    # | |\/| || | | |  | |    | |   | |     / _` || || || '_ \  / _` | / __|| |/ // __|
    # | |  | |\ \/' /  | |    | |   | \__/\| (_| || || || |_) || (_| || (__ |   < \__ \
    # \_|  |_/ \_/\_\  \_/    \_/    \____/ \__,_||_||_||_.__/  \__,_| \___||_|\_\|___/
    # ===================================================================================

    def _on_connect(self, client: MQTTClient, flags, rc, properties):
        if rc == 0:
            MXLOG_DEBUG(f'{PrintTag.CONNECT} Connect to Host: {self._ip}:{self._port}')
        else:
            MXLOG_DEBUG(f'{PrintTag.ERROR} Bad connection... Returned code: {rc}', 'red')

    def _on_disconnect(self, client: MQTTClient, packet, exc=None):
        self.next_state = ThingState.BROKER_LOST
        MXLOG_DEBUG(f'{PrintTag.DISCONNECT} Disconnect from broker {self._ip}:{self._port}')

    def _on_subscribe(self, client: MQTTClient, mid, qos, properties):
        pass

    def _on_unsubscribe(self, client: MQTTClient, mid, qos):
        pass

    # def _on_publish(self, client: MQTTClient, userdata: MQTTMessage, mid):
    #     pass

    async def _on_message(self, client: MQTTClient, topic, payload, qos, properties):
        # topic, payload = decode_MQTT_message(msg)
        self._print_packet(topic=topic, payload=payload, direction=Direction.RECEIVED, mode=self._log_mode)
        msg = encode_MQTT_message(topic, payload)

        protocol = MXProtocolType.get(topic)
        if protocol in [
            MXProtocolType.Base.MT_REQUEST_REGISTER_INFO,
            MXProtocolType.Base.MT_REQUEST_UNREGISTER,
            MXProtocolType.Base.MT_RESULT_REGISTER,
            MXProtocolType.Base.MT_RESULT_UNREGISTER,
            MXProtocolType.Base.MT_RESULT_BINARY_VALUE,
            MXProtocolType.Base.MT_EXECUTE,
            MXProtocolType.Base.MT_IN_EXECUTE,
            MXProtocolType.WebClient.ME_RESULT_HOME,
        ]:
            await self._receive_queue[protocol].put(msg)
        else:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Unexpected topic! topic: {topic}')

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def function_list(self) -> List[MXFunction]:
        return sorted(
            [service for service in self._thing_data.service_list if isinstance(service, MXFunction) and not service.name.startswith('__')],
            key=lambda x: x.name,
        )

    @property
    def value_list(self) -> List[MXValue]:
        return sorted([service for service in self._thing_data.service_list if isinstance(service, MXValue)], key=lambda x: x.name)

    @property
    def service_list(self) -> List[MXService]:
        return sorted(self._thing_data.service_list, key=lambda x: x.name)

    @property
    def name(self) -> str:
        return self._thing_data.name

    @property
    def nick_name(self) -> str:
        return self._thing_data.nick_name

    @property
    def middleware_name(self) -> str:
        return self._thing_data.middleware_name

    @property
    def last_alive_time(self) -> float:
        return self._thing_data.last_alive_time

    @property
    def alive_cycle(self) -> float:
        return self._thing_data.alive_cycle

    @property
    def subscribed_topic_set(self) -> Set[str]:
        return self._thing_data.subscribed_topic_set

    @property
    def is_super(self) -> bool:
        return self._thing_data.is_super

    @property
    def is_parallel(self) -> bool:
        return self._thing_data.is_parallel

    @property
    def is_builtin(self) -> bool:
        return self._thing_data.is_builtin

    @property
    def is_manager(self) -> bool:
        return self._thing_data.is_manager

    @property
    def is_matter(self) -> bool:
        return self._thing_data.is_matter

    @property
    def current_state(self):
        return self._state

    def get_log_mode(self) -> MXPrintMode:
        return self._log_mode

    def get_log_enable(self) -> bool:
        return self._log_enable

    def get_log_name(self) -> str:
        return self._log_enable

    def get_ip(self) -> str:
        return self._ip

    def get_port(self) -> int:
        return self._port

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @name.setter
    def name(self, name: str) -> None:
        self._thing_data.name = name
        for service in self._thing_data.service_list:
            service.thing_name = name

    @nick_name.setter
    def nick_name(self, nick_name: str) -> None:
        self._thing_data.nick_name = nick_name

    @middleware_name.setter
    def middleware_name(self, middleware_name: str) -> None:
        self._thing_data.middleware_name = middleware_name
        for service in self._thing_data.service_list:
            service.middleware_name = middleware_name

    @last_alive_time.setter
    def last_alive_time(self, last_alive_time: float) -> None:
        self._thing_data.last_alive_time = last_alive_time

    @alive_cycle.setter
    def alive_cycle(self, alive_cycle: float) -> None:
        self._thing_data.alive_cycle = alive_cycle

    @subscribed_topic_set.setter
    def subscribe_topic_set(self, subscribe_topic_set: Set[str]) -> None:
        self._thing_data.subscribed_topic_set = subscribe_topic_set

    @is_super.setter
    def is_super(self, is_super: bool) -> bool:
        self._thing_data.is_super = is_super

    @is_parallel.setter
    def is_parallel(self, is_parallel: bool) -> bool:
        self._thing_data.is_parallel = is_parallel

    @is_builtin.setter
    def is_builtin(self, is_builtin: bool) -> bool:
        self._thing_data.is_builtin = is_builtin

    @is_manager.setter
    def is_manager(self, is_manager: bool) -> bool:
        self._thing_data.is_manager = is_manager

    @is_matter.setter
    def is_matter(self, is_matter: bool) -> bool:
        self._thing_data.is_matter = is_matter

    @current_state.setter
    def next_state(self, state: ThingState):
        self._state = state
        MXLOG_DEBUG(f'Current state: {self._state}', 'cyan')

    def set_log_mode(self, mode: MXPrintMode) -> None:
        self._log_mode = mode

    def set_log_enable(self, enable: bool) -> None:
        self._log_enable = enable

    def set_log_name(self, name: str) -> None:
        self._log_name = name

    def set_ip(self, ip: str) -> None:
        self._ip = ip

    def set_port(self, port: int) -> None:
        self._port = port



from enum import Enum, auto


class MXErrorCode(Enum):
    NO_ERROR = 0
    FAIL = -1
    TIMEOUT = -2
    NO_PARALLEL = -3
    DUPLICATE = -4
    TARGET_NOT_FOUND = -5
    REQUESTER_NOT_FOUND = -6
    INVALID_REQUEST = -7
    INVALID_DATA = -8
    INVALID_DESTINATION = -9
    TOO_LONG_IDENTIFIER = -10
    EXECUTE_FAIL = -11
    EXECUTE_TIMEOUT = -12
    EXECUTE_NO_PARALLEL = -13
    EXECUTE_DUPLICATE = -14
    EXECUTE_ARG_FAIL = -15
    VALUE_ERROR = -16
    TYPE_ERROR = -18
    NOT_FOUND_ERROR = -19
    CONNECTION_ERROR = -20
    NOT_SUPPORTED = -21

    UNDEFINED = 'undefined'

    @classmethod
    def to_mx_error_code(cls, type: int):
        if type is not None:
            for mx_error_code in MXErrorCode:
                if mx_error_code.value == type:
                    return mx_error_code
        else:
            return MXErrorCode.get(type)

    @classmethod
    def get(cls, type: int) -> 'MXErrorCode':
        try:
            if type is not None:
                for mx_error_type in MXErrorCode:
                    if mx_error_type.value == type:
                        return mx_error_type
                return cls.UNDEFINED
            else:
                return MXErrorCode.get(type)
        except Exception:
            return cls.UNDEFINED

    def __str__(self):
        return self.name


class MXError(Exception):
    def __init__(self, error_code: MXErrorCode, *args: object, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)
        self._error_code: MXErrorCode = error_code


class MXFailError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.FAIL, *args, **kwargs)


class MXTimeoutError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.TIMEOUT, *args, **kwargs)
        self._error_code = MXErrorCode.TIMEOUT


class MXNoParallelError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.NO_PARALLEL, *args, **kwargs)


class MXDuplicatedError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.DUPLICATE, *args, **kwargs)


class MXNotSupportedError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.NOT_SUPPORTED, *args, **kwargs)


class MXInvalidRequestError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.INVALID_REQUEST, *args, **kwargs)


class MXValueError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.VALUE_ERROR, *args, **kwargs)


class MXTypeError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.TYPE_ERROR, *args, **kwargs)


class MXNotFoundError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.NOT_FOUND_ERROR, *args, **kwargs)


class MXConnectionError(MXError):
    def __init__(self, *args: object, **kwargs: dict) -> None:
        super().__init__(MXErrorCode.CONNECTION_ERROR, *args, **kwargs)



from big_thing_py.super.super_request import *
from big_thing_py.super.super_mqtt_message import *
import threading
import random
from func_timeout import FunctionTimedOut


class MXSuperFunction(MXFunction):
    '''
    [Whole super service structure]
    super_thing┬── super_service1┬── sub_service1┬────── target_thing1(at Middleware1)
               │                  │               ╲
               │                  │                ╲──── target_thing2(at Middleware1)
               │                  │                 ╲
               │                  │                  ╲── target_thing3(at Middleware2)
               │                  │                   ╲
               │                  ├── sub_service2      ╲ ....
               │                  │
               │                  ├── ...
               │
               │
               └── super_service2 ── sub_service1(target Thing1, target Middleware2)
    '''

    def __init__(
        self,
        func: Callable,
        return_type: MXType,
        name: str = '',
        tag_list: List[MXTag] = [],
        energy: float = 0,
        desc: str = '',
        thing_name: str = '',
        middleware_name: str = '',
        arg_list: List[MXArgument] = [],
        exec_time: float = 0,
        timeout: float = 0,
        range_type: MXRangeType = MXRangeType.SINGLE,
    ) -> None:
        super().__init__(
            func=func,
            return_type=return_type,
            name=name,
            tag_list=tag_list,
            energy=energy,
            desc=desc,
            thing_name=thing_name,
            middleware_name=middleware_name,
            arg_list=arg_list,
            exec_time=exec_time,
            timeout=timeout,
            range_type=range_type,
        )

        # 사용자가 super thing코드에 명세한 sub_service 조건에 맞는 sub_service 종류. 해당 조건에 맞는 sub_service 여러 디바이스에 여러 개 존재할 수 있다.
        # super service에 명세된 req() 라인 수 만큼 존재한다.
        self._sub_service_request_list: List[MXSubServiceRequest] = []

        # SuperService 안에 있는 req()들에 대한 정보가 스캔되었는지 여부. 스캔된 req()들에 대한 정보는 _sub_service_type_list에 저장된다.
        self._is_scanned: bool = False

        # scheduling에 대한 결과를 super thing이 넣어주기 위해 존재함
        #     super_service_request_key: str = make_super_request_key(requester_middleware, scenario)
        #     super_service_request_key: {requester_middleware}@{scenario}, value: MXSuperScheduleRequest
        self._temporary_scheduling_table: Dict[str, MXSuperScheduleRequest] = {}

        # scheduling이 끝나면 해당 scenario에 대해 어떤 sub_service_type에 대해 thing을 선택할 것인지 정해진다.
        # 이렇게 하는 이유는 scenario를 반복해서 돌릴 때마다 다른 thing을 선택하는 것이 아니라 전에 선택했던 thing에 있는 service를 실행하기 위함이다.
        #     super_service_request_key: str = make_super_request_key(requester_middleware, scenario)
        #     super_service_request_key: {requester_middleware}@{scenario}, value: MXSuperExecuteRequest
        self._mapping_table: Dict[str, MXSuperExecuteRequest] = {}

        self._schedule_running: bool = False

        if not inspect.ismethod(func):
            raise MXTypeError('self._func must be a instance method')

    def __eq__(self, o: 'MXSuperFunction') -> bool:
        instance_check = isinstance(o, MXSuperFunction)

        return super().__eq__(o) and instance_check

    def __getstate__(self):
        state = super().__getstate__()

        state['_sub_service_request_list'] = self._sub_service_request_list
        state['_is_scanned'] = self._is_scanned
        state['_temporary_scheduling_table'] = self._temporary_scheduling_table
        state['_mapping_table'] = self._mapping_table

        del state['_schedule_running']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._sub_service_request_list = state['_sub_service_request_list']
        self._is_scanned = state['_is_scanned']
        self._temporary_scheduling_table = state['_temporary_scheduling_table']
        self._mapping_table = state['_mapping_table']

        self._schedule_running = False

    def _add_sub_service_request(self, sub_service_request: MXSubServiceRequest):
        sub_service_request._sub_service_request_order = len(self._sub_service_request_list)
        self._sub_service_request_list.append(sub_service_request)

    def _remove_sub_service_request(self, sub_service_request: MXSubServiceRequest):
        if sub_service_request in self._sub_service_request_list:
            self._sub_service_request_list.remove(sub_service_request)

    def _request_subschedule(self, subschedule_request: MXSubScheduleRequest) -> None:
        subschedule_request.timer_start()
        subschedule_msg = subschedule_request.trigger_msg

        # 필요한 토픽 구독과 발행을 한번에 수행한다.
        if not isinstance(subschedule_request, MXSubScheduleRequest):
            raise MXTypeError(f'Invalid type of subschedule_request: {type(subschedule_request)}')

        self._send_SM_SCHEDULE(subschedule_msg)

    def _select_target_sub_service_by_range_type(
        self, super_schedule_request: MXSuperScheduleRequest, single_select_func: Callable
    ) -> MXSubScheduleRequest:
        for sub_service_request in self._sub_service_request_list:
            sub_service_request_range_type = sub_service_request._sub_service_type.range_type
            sub_service_request_key = make_sub_service_request_key(
                sub_service_name=sub_service_request._sub_service_type.name,
                sub_service_request_order=sub_service_request._sub_service_request_order,
            )
            target_sub_service_request = super_schedule_request._sub_service_request_table[sub_service_request_key]

            if sub_service_request_range_type == MXRangeType.ALL:
                for candidate_request in target_sub_service_request._candidate_request_list:
                    if candidate_request._status == MXScheduleStatus.CHECK:
                        candidate_request._status = MXScheduleStatus.SELECT
            elif sub_service_request_range_type == MXRangeType.SINGLE:
                while True:
                    selected_request = single_select_func(target_sub_service_request._candidate_request_list)
                    if selected_request._status == MXScheduleStatus.CHECK:
                        selected_request._status = MXScheduleStatus.SELECT
                        break

    def _subschedule_parallel(
        self,
        super_schedule_request: MXSuperScheduleRequest,
        target_schedule_status: MXScheduleStatus,
        result_schedule_status: MXScheduleStatus,
    ) -> List[MXSubScheduleRequest]:
        subschedule_start_time = get_current_datetime()

        # 병렬로 subschedule 요청을 보낸다.
        for sub_service_request in super_schedule_request._sub_service_request_table.values():
            candidate_subschedule_request_list = [
                candidate_request
                for candidate_request in sub_service_request._candidate_request_list
                if candidate_request._status == target_schedule_status
            ]

            for candidate_subschedule_request in candidate_subschedule_request_list:
                # 미들웨어에게 현재 요청인 check 인지 confirm인지 알려주기 위해 trigger_msg에 status를 추가함.
                candidate_subschedule_request.trigger_msg.status = result_schedule_status
                candidate_subschedule_msg = candidate_subschedule_request.trigger_msg
                MXLOG_DEBUG(
                    f'[SUB_SCHEDULE {candidate_subschedule_msg._status.value.upper()} START] '
                    f'{candidate_subschedule_msg.sub_service_name}|{candidate_subschedule_msg.target_thing_name}|'
                    f'{candidate_subschedule_msg.target_middleware_name}',
                    'cyan',
                )
                self._request_subschedule(candidate_subschedule_request)

            # 해당 super service의 sub_service들에게 subschedule 요청을 보낸 후, 모든 subschedule 결과를 받을 때까지 기다린다.
            for candidate_subschedule_request in candidate_subschedule_request_list:
                candidate_subschedule_request.result_msg = candidate_subschedule_request.get_result_msg()
                candidate_subschedule_request._status = result_schedule_status

                candidate_subschedule_request.timer_end()
                MXLOG_DEBUG(
                    f'[SUB_SCHEDULE {candidate_subschedule_request._status.value.upper()} END] '
                    f'{candidate_subschedule_request.trigger_msg.sub_service_name}|{candidate_subschedule_request.trigger_msg.target_thing_name}|'
                    f'{candidate_subschedule_request.trigger_msg.target_middleware_name}, duration: {candidate_subschedule_request.duration():.4f} Sec',
                    'cyan',
                )

                # 굳이 구독 해제를 할 필요가 없다. 재호출될 가능성이 높음.
                # subschedule_result_msg: MXSubScheduleResultMessage = candidate_subschedule_request._result_msg
                # self._unsubscribe_queue.put(subschedule_result_msg.topic)
                # while not self._unsubscribe_queue.empty():
                #     time.sleep(0.001)

        super_schedule_request._status = result_schedule_status
        if super_schedule_request._status == MXScheduleStatus.CHECK:
            super_schedule_request._check_duration = get_current_datetime() - subschedule_start_time
            MXLOG_DEBUG(
                f'[SUPER_SCHEDULE CHECK END] '
                f'{super_schedule_request.trigger_msg.super_service_name}|{super_schedule_request.trigger_msg.super_thing_name}|'
                f'{super_schedule_request.trigger_msg.super_middleware_name}, duration: {super_schedule_request.confirm_duration():.4f} Sec',
                'green',
            )
        elif super_schedule_request._status == MXScheduleStatus.CONFIRM:
            super_schedule_request._confirm_duration = get_current_datetime() - subschedule_start_time
            MXLOG_DEBUG(
                f'[SUPER_SCHEDULE CONFIRM END] '
                f'{super_schedule_request.trigger_msg.super_service_name}|{super_schedule_request.trigger_msg.super_thing_name}|'
                f'{super_schedule_request.trigger_msg.super_middleware_name}, duration: {super_schedule_request.confirm_duration():.4f} Sec',
                'green',
            )

    def _check_parallel(self, super_schedule_request: MXSuperScheduleRequest) -> List[MXSubScheduleRequest]:
        self._subschedule_parallel(
            super_schedule_request=super_schedule_request,
            target_schedule_status=MXScheduleStatus.INIT,
            result_schedule_status=MXScheduleStatus.CHECK,
        )

    def _confirm_parallel(self, super_schedule_request: MXSuperScheduleRequest) -> bool:
        self._subschedule_parallel(
            super_schedule_request=super_schedule_request,
            target_schedule_status=MXScheduleStatus.SELECT,
            result_schedule_status=MXScheduleStatus.CONFIRM,
        )

    def _check_sub_service_request_confirm(self, super_schedule_request: MXSuperScheduleRequest) -> bool:
        # 스케쥴링 과정에서 super service의 sub_service_request 중 confirm 된 request가 하나도 없는 sub_service_request가 존재한다면 에러를 발생시킨다.
        for sub_service_request in super_schedule_request._sub_service_request_table.values():
            for candidate_request in sub_service_request._candidate_request_list:
                if candidate_request._status == MXScheduleStatus.CONFIRM:
                    break
            else:
                return False
        else:
            return True

    def _check_confirm_parallel(self, super_schedule_request: MXSuperScheduleRequest) -> bool:
        self._check_parallel(super_schedule_request=super_schedule_request)
        self._select_target_sub_service_by_range_type(super_schedule_request=super_schedule_request, single_select_func=lambda x: x[0])
        self._confirm_parallel(super_schedule_request=super_schedule_request)

        return True

    def _print_schedule_result(self, scenario_name: str, requester_middleware_name: str):
        schedule_result_string = '\n[SCHEDULE RESULT] ============================================\n'
        super_request_key = make_super_request_key(scenario_name=scenario_name, requester_middleware_name=requester_middleware_name)
        for super_k, super_execute_req in self._mapping_table.items():
            if super_k != super_request_key:
                continue
            schedule_result_string += f'super_key: {super_k} -> super_service: {self._name}\n'
            for sub_k, sub_service_req in super_execute_req._sub_service_request_table.items():
                schedule_result_string += (
                    ' ' * 4
                    + f'sub_key: {sub_k} -> sub_service: {sub_service_req._sub_service_type.name}|{sub_service_req._sub_service_request_order}\n'
                )
                for target_req in sub_service_req._target_request_list:
                    schedule_result_string += (
                        ' ' * 8 + f'target: {target_req.trigger_msg.sub_service_name}|{target_req.trigger_msg.target_middleware_name}\n'
                    )
        schedule_result_string += '==============================================================\n'
        MXLOG_DEBUG(schedule_result_string, 'green')

    def _super_schedule_wrapper(
        self,
        super_schedule_request: MXSuperScheduleRequest,
        hierarchical_service_table: Dict[str, List[MXService]],
        timeout: float = 1000,
    ):
        '''
        super service의 하위 함수에 대한 정보를 추출한다.
        service_list 구조는

        ============================================
        super_service -> sub_function_type_list
                       -> sub_function_list
        ============================================

        로 이루어져있다.
        sub_function_type_list과 sub_function_list는 독립적인 공간을 가진다.
        super_service 내부에 req함수 가 존재하여 사용자가 요청하고 싶은 sub_service이 명세되어있는데 여기서 명세되어지는
        sub_service은 실제 타겟 sub_service이 아닌 sub_service_type이다. 실제 sub_service 정보는 middleware로 부터 받은
        service_list를 통해 추출한다. 그리고 해당 정보는 super_service의 sub_service_list에 저장된다.
        '''
        if not isinstance(super_schedule_request, MXSuperScheduleRequest):
            raise MXTypeError(f'[{get_current_function_name()}] Wrong Request type: {type(super_schedule_request)}')

        super_schedule_msg = super_schedule_request.trigger_msg
        requester_middleware_name = super_schedule_msg.requester_middleware_name
        scenario = super_schedule_msg.scenario
        super_service_request_key = make_super_request_key(scenario_name=scenario, requester_middleware_name=requester_middleware_name)

        MXLOG_DEBUG(f'[SUPER_SCHEDULE START] {self._name} by scenario {scenario}.', 'green')

        try:
            super_schedule_request._running = True
            self._schedule_running = True
            error = MXErrorCode.NO_ERROR

            # 후보 sub_service들에 대해서 스케쥴링을 진행한다.
            super_schedule_request.generate_sub_service_schedule_request(self._sub_service_request_list, hierarchical_service_table['functions'])

            # func_timeout(timeout, self._check_confirm_parallel, args=(super_schedule_request, ))
            current_thread = threading.current_thread()
            self._run_with_timeout(
                timeout=timeout,
                func=self._check_confirm_parallel,
                name=current_thread.name,
                user_data=dict(scenario=super_schedule_msg.scenario, requester_middleware=requester_middleware_name),
                args=(super_schedule_request,),
            )

            # 만약 현재 super service의 sub_service_list가 비어있는 경우 에러를 발생시킨다.
            if len(super_schedule_request._sub_service_request_table) == 0 and len(self._sub_service_request_list) != 0:
                raise Exception(f'No target_sub_service_list found in {self._name} super service')
        except KeyboardInterrupt as e:
            print_error(e)
            MXLOG_DEBUG('Function scheduling exit by user', 'red')
            raise e
        except FunctionTimedOut as e:
            MXLOG_DEBUG(
                f'[SUPER_SCHEDULE TIMEOUT] {super_schedule_request._action_type.value} of {self._name}'
                'by scenario {requester_middleware_name}|{scenario} timeout...',
                'red',
            )

            error = MXErrorCode.TIMEOUT
            return False
        except Exception as e:
            print_error(e)
            MXLOG_DEBUG(
                f'[SUPER_SCHEDULE FAILED] {super_schedule_request._action_type.value} of {self._name}'
                'by scenario {requester_middleware_name}|{scenario} failed...',
                'red',
            )

            error = MXErrorCode.FAIL
            return False
        else:
            if not self._check_sub_service_request_confirm(super_schedule_request):
                return False

            super_service_execute_request = MXSuperExecuteRequest(super_schedule_request=super_schedule_request)
            self._mapping_table[super_service_request_key] = super_service_execute_request

            error = error = MXErrorCode.NO_ERROR
            return True
        finally:
            self._send_SM_RESULT_SCHEDULE(
                super_service_name=super_schedule_msg.super_service_name,
                super_thing_name=super_schedule_msg.super_thing_name,
                super_middleware_name=super_schedule_msg.super_middleware_name,
                requester_middleware_name=requester_middleware_name,
                scenario=scenario,
                error=error,
            )

            super_schedule_request.timer_end()
            super_schedule_request._running = False
            self._schedule_running = False
            self._temporary_scheduling_table.pop(super_service_request_key)
            MXLOG_DEBUG(
                f'[SUPER_SCHEDULE END] {self._name} by scenario {scenario}. duration: {super_schedule_request.duration():.4f} Sec',
                'green',
            )
            self._print_schedule_result(scenario_name=scenario, requester_middleware_name=requester_middleware_name)

    def _super_execute_wrapper(self, super_service_execute_request: MXSuperExecuteRequest):
        '''
        super service의 하위 함수에 대한 정보를 추출한다.
        service_list 구조는

        ============================================
        super_service -> sub_function_type_list
                       -> sub_function_list
        ============================================

        로 이루어져있다.
        sub_function_type_list과 sub_function_list는 독립적인 공간을 가진다.
        super_service 내부에 req함수 가 존재하여 사용자가 요청하고 싶은 sub_service이 명세되어있는데 여기서 명세되어지는
        sub_service은 실제 타겟 sub_service이 아닌 sub_service_type이다. 실제 sub_service 정보는 middleware로 부터 받은
        service_list를 통해 추출한다. 그리고 해당 정보는 super_service의 sub_service_list에 저장된다.
        '''

        if not isinstance(super_service_execute_request, MXSuperExecuteRequest):
            raise MXTypeError(f'[{get_current_function_name()}] Wrong Request type: {type(super_service_execute_request)}')

        super_execute_msg = super_service_execute_request.trigger_msg
        requester_middleware_name = super_execute_msg.requester_middleware_name
        scenario = super_execute_msg.scenario

        MXLOG_DEBUG(f'[SUPER_EXECUTE START] {self._name} by scenario {scenario}.', 'green')

        try:
            super_service_execute_request._running = True
            self._running = True
            error = MXErrorCode.NO_ERROR

            # super service argument check
            arguments = list(super_execute_msg.tuple_arguments())
            for i, (arg, real_arg) in enumerate(zip(self._arg_list, arguments)):
                if arg.get_type in [MXType.INTEGER, MXType.DOUBLE] and MXType.get(type(real_arg)) in [
                    MXType.INTEGER,
                    MXType.DOUBLE,
                ]:
                    arguments[i] = float(real_arg)
                elif arg.get_type != MXType.get(type(real_arg)):
                    raise Exception(f'Argument type is not matched: {arg.name}: {arg.get_type} != {MXType.get(real_arg)}')

            if self._timeout:
                # self._return_value = func_timeout(self._timeout, self._func, args=(*tuple(arguments), ))
                current_thread = threading.current_thread()
                self._return_value = self._run_with_timeout(
                    timeout=self._timeout,
                    func=self._func,
                    name=current_thread.name,
                    user_data=dict(scenario=super_execute_msg.scenario, requester_middleware=requester_middleware_name),
                    args=(*tuple(arguments),),
                )
            else:
                self._return_value = self._func(*tuple(arguments))
        except KeyboardInterrupt as e:
            print_error(e)
            MXLOG_DEBUG('Function execution exit by user', 'red')
            raise e
        except FunctionTimedOut as e:
            MXLOG_DEBUG(
                f'[SUPER_EXECUTE TIMEOUT] {super_service_execute_request._action_type.value} of {self._name} \
                    by scenario {requester_middleware_name}|{scenario} timeout...',
                'red',
            )

            # Timeout이 발생했다고 해서 mapping_table에서 삭제하지는 않는다.
            # if super_service_execute_request._action_type == MXActionType.SUPER_EXECUTE:
            #     self._mapping_table.pop(super_service_request_key)

            error = MXErrorCode.TIMEOUT
            self._return_value = None
            return False
        except Exception as e:
            print_error(e)
            MXLOG_DEBUG(
                f'[SUB_EXECUTE FAILED] {super_service_execute_request._action_type.value} of {self._name} \
                by scenario {requester_middleware_name}|{scenario} failed...',
                'red',
            )

            # Exception이 발생했다고 해서 mapping_table에서 삭제하지는 않는다.
            # if super_service_execute_request._action_type == MXActionType.SUPER_EXECUTE:
            #     self._mapping_table.pop(super_service_request_key)

            error = MXErrorCode.FAIL
            self._return_value = None
            return False
        else:
            error = MXErrorCode.NO_ERROR
            return True
        finally:
            self._send_SM_RESULT_EXECUTE(
                super_service_name=super_execute_msg.super_service_name,
                super_thing_name=super_execute_msg.super_thing_name,
                super_middleware_name=super_execute_msg.super_middleware_name,
                requester_middleware_name=requester_middleware_name,
                scenario=scenario,
                error=error,
            )

            super_service_execute_request.timer_end()
            super_service_execute_request._running = False
            self._running = False
            self._running_scenario_list.remove(super_execute_msg.scenario)
            MXLOG_DEBUG(
                f'[SUPER_EXECUTE END] {self._name} by scenario {scenario}. duration: {super_service_execute_request.duration():.4f} Sec',
                'green',
            )

    def start_super_schedule_thread(
        self,
        super_schedule_msg: MXSuperScheduleMessage,
        hierarchical_service_table: Dict[str, List[MXService]],
        timeout: float = 1000,
    ) -> MXThread:
        super_service_request_key = make_super_request_key(
            scenario_name=super_schedule_msg.scenario,
            requester_middleware_name=super_schedule_msg.requester_middleware_name,
        )
        if super_service_request_key in self._temporary_scheduling_table:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super request with key: {super_service_request_key} is currently in the scheduling phase!',
                'red',
            )
            return False

        super_schedule_result_msg = MXSuperScheduleResultMessage(
            super_service_name=self._name,
            super_thing_name=self._thing_name,
            super_middleware_name=self._middleware_name,
            requester_middleware_name=super_schedule_msg.requester_middleware_name,
            scenario=super_schedule_msg.scenario,
        )
        super_schedule_request = MXSuperScheduleRequest(trigger_msg=super_schedule_msg, result_msg=super_schedule_result_msg)
        super_service_request_key = make_super_request_key(
            scenario_name=super_schedule_msg.scenario,
            requester_middleware_name=super_schedule_msg.requester_middleware_name,
        )
        self._temporary_scheduling_table[super_service_request_key] = super_schedule_request
        super_schedule_request.timer_start()

        super_schedule_thread = MXThread(
            target=self._super_schedule_wrapper,
            name=f'{self._func.__name__}_{super_schedule_request._action_type.value}_thread',
            daemon=True,
            args=(
                super_schedule_request,
                hierarchical_service_table,
                timeout,
            ),
        )
        super_schedule_thread.start()

        return super_schedule_thread

    def start_super_execute_thread(self, super_execute_msg: MXSuperExecuteMessage, SUPER_SERVICE_REQUEST_KEY_TABLE: Dict[str, List[str]]) -> MXThread:
        super_service_request_key = make_super_request_key(
            scenario_name=super_execute_msg.scenario,
            requester_middleware_name=super_execute_msg.requester_middleware_name,
        )
        if super_service_request_key in self._temporary_scheduling_table:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super request with key: {super_service_request_key} is currently in the scheduling phase!',
                'red',
            )
            return False

        super_service_execute_request = self._mapping_table[super_service_request_key]
        if super_service_execute_request._running:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Execute of Super request key: {super_service_request_key} is already executed!',
                'red',
            )
            return False

        super_service_execute_request.trigger_msg = super_execute_msg
        super_service_execute_request.result_msg = MXSuperExecuteResultMessage(
            super_service_name=self._name,
            super_thing_name=self._thing_name,
            super_middleware_name=self._middleware_name,
            requester_middleware_name=super_execute_msg.requester_middleware_name,
            scenario=super_execute_msg.scenario,
            return_type=self._return_type,
        )
        self._running_scenario_list.append(super_execute_msg.scenario)
        super_service_execute_request.timer_start()

        thread_name = f'{self._func.__name__}_{super_execute_msg.scenario}_thread'
        super_execute_thread = MXThread(target=self._super_execute_wrapper, name=thread_name, daemon=True, args=(super_service_execute_request,))
        super_execute_thread.start()

        sub_service_request_key_list = list(super_service_execute_request._sub_service_request_table)
        SUPER_SERVICE_REQUEST_KEY_TABLE[super_service_request_key] = sub_service_request_key_list

        return super_execute_thread

    # FIXME: hardcoding 된 부분 수정하기 (DUMMY_ARG_)
    def add_sub_service_request_info(
        self,
        sub_service_name: str,
        arg_list: Union[Tuple[MXArgument], Tuple],
        tag_list: List[MXTag],
        return_type: MXType,
        range_type: MXRangeType,
    ):
        mx_arg_list = [
            (
                arg
                if isinstance(arg, MXArgument)
                else MXArgument(name=f'DUMMY_ARG_{i}', bound=(-sys.maxsize - 1, sys.maxsize), type=MXType.get(type(arg)))
            )
            for i, arg in enumerate(arg_list)
        ]
        sub_service_request_order = len(self._sub_service_request_list)
        sub_service_type = MXFunction(
            name=sub_service_name,
            func=dummy_func(arg_list=mx_arg_list),
            return_type=return_type,
            arg_list=mx_arg_list,
            tag_list=tag_list,
            range_type=range_type,
        )
        sub_service_request = MXSubServiceRequest(sub_service_type=sub_service_type, sub_service_request_order=sub_service_request_order)
        self._add_sub_service_request(sub_service_request)

        MXLOG_DEBUG(f'sub_service: {sub_service_name}:{sub_service_request_order}', 'green')
        return True

    def put_subschedule_result(
        self,
        super_service_request_key: str,
        sub_service_request_key: str,
        subschedule_result_msg: MXSubScheduleResultMessage,
    ) -> bool:
        if super_service_request_key not in self._temporary_scheduling_table:
            MXLOG_DEBUG(
                f'Could not find key {super_service_request_key} in temporary_scheduling_table... This key is not mine!',
                'yellow',
            )
            return False

        super_schedule_request = self._temporary_scheduling_table[super_service_request_key]
        result = super_schedule_request.put_sub_service_schedule_result_msg(sub_service_request_key, subschedule_result_msg)
        return result

    def put_sub_service_execute_result(
        self,
        super_service_request_key: str,
        sub_service_request_key: str,
        sub_service_execute_result_msg: MXSubExecuteResultMessage,
    ) -> bool:
        if super_service_request_key not in self._mapping_table:
            MXLOG_DEBUG(f'Could not find key {super_service_request_key} in mapping_table... This key is not mine!', 'yellow')
            return False

        super_service_execute_request = self._mapping_table[super_service_request_key]
        result = super_service_execute_request.put_sub_service_execute_result_msg(sub_service_request_key, sub_service_execute_result_msg)
        return result

    def _send_SM_SCHEDULE(self, subschedule_msg: MXSubScheduleMessage) -> None:
        subschedule_mqtt_msg = subschedule_msg.mqtt_message()
        self._publish_queue.put(subschedule_mqtt_msg)

    def _send_SM_EXECUTE(self, sub_service_execute_msg: MXSubExecuteMessage) -> None:
        sub_service_execute_mqtt_msg = sub_service_execute_msg.mqtt_message()
        self._publish_queue.put(sub_service_execute_mqtt_msg)

    def _send_SM_RESULT_SCHEDULE(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        error: MXErrorCode,
    ) -> None:
        super_schedule_result_msg = self.generate_super_schedule_result_message(
            super_service_name=super_service_name,
            super_thing_name=super_thing_name,
            super_middleware_name=super_middleware_name,
            requester_middleware_name=requester_middleware_name,
            scenario=scenario,
            error=error,
        )
        super_schedule_result_mqtt_msg = super_schedule_result_msg.mqtt_message()
        self._publish_queue.put(super_schedule_result_mqtt_msg)

    def _send_SM_RESULT_EXECUTE(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        error: MXErrorCode,
    ):
        super_execute_result_msg = self.generate_super_execute_result_message(
            super_service_name=super_service_name,
            super_thing_name=super_thing_name,
            super_middleware_name=super_middleware_name,
            requester_middleware_name=requester_middleware_name,
            scenario=scenario,
            error=error,
        )
        super_execute_result_mqtt_msg = super_execute_result_msg.mqtt_message()
        self._publish_queue.put(super_execute_result_mqtt_msg)

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def generate_super_schedule_result_message(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        error: MXErrorCode,
    ) -> MXSuperScheduleResultMessage:
        super_schedule_result_msg = MXSuperScheduleResultMessage(
            super_service_name=super_service_name,
            super_thing_name=super_thing_name,
            super_middleware_name=super_middleware_name,
            requester_middleware_name=requester_middleware_name,
            scenario=scenario,
            error=error,
        )
        return super_schedule_result_msg

    def generate_super_execute_result_message(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        error: MXErrorCode,
    ) -> MXSuperExecuteResultMessage:
        super_execute_result_msg = MXSuperExecuteResultMessage(
            super_service_name=super_service_name,
            super_thing_name=super_thing_name,
            super_middleware_name=super_middleware_name,
            requester_middleware_name=requester_middleware_name,
            scenario=scenario,
            return_type=self._return_type,
            return_value=self._return_value,
            error=error,
        )
        return super_execute_result_msg

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    def get_sub_service_type_list(self) -> List[MXSubServiceRequest]:
        return self._sub_service_request_list

    def get_is_scanned(self) -> bool:
        return self._is_scanned

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    def set_sub_service_type_list(self, target_sub_service_request_list: List[MXSubServiceRequest]) -> None:
        self._sub_service_request_list = target_sub_service_request_list

    def set_is_scanned(self, is_scanned: bool) -> None:
        self._is_scanned = is_scanned



import base64
import re


def file_to_base64(file_path: str) -> str:
    with open(file_path, 'rb') as f:
        return base64.b64encode(f.read()).decode()


def base64_to_file(base64_string: str, file_path: str) -> None:
    with open(file_path, 'wb') as f:
        f.write(base64.b64decode(base64_string))


def string_to_base64(string: str) -> str:
    return base64.b64encode(string.encode()).decode()


def base64_to_string(base64_string: str) -> str:
    return base64.b64decode(base64_string).decode()


def is_base64(base64_str: str) -> bool:
    '''
    NOTE: The minimum length set to 24 characters is chosen based on common use cases,
          such as encoded hashes or small files, which typically result in a Base64
          string longer than 24 characters. This length helps reduce the chance of
          mistaking regular short strings for Base64 encoded data.
    '''
    if len(base64_str) < 24 or len(base64_str) % 4 != 0:
        return False

    pattern = re.compile(r'^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$')
    if bool(pattern.match(base64_str)):
        try:
            base64.b64decode(base64_str, validate=True)
            return True
        except Exception:
            return False
    else:
        return False


if __name__ == '__main__':
    print(is_base64('This is test string.'))



from typing import *
from typing_extensions import override, overload
from termcolor import *
from abc import *
from enum import Enum, auto


EMPTY_JSON = {}
THREAD_TIME_OUT = 0.01


class TimeFormat(Enum):
    DATETIME1 = '%Y-%m-%d %H:%M:%S'
    DATETIME2 = '%Y%m%d_%H%M%S'
    DATE = '%Y-%m-%d'
    TIME = '%H:%M:%S'
    UNIXTIME = 'unixtime'


class MXPrintMode(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    UNDEFINED = auto()
    FULL = auto()
    ABBR = auto()
    SKIP = auto()

    @classmethod
    def get(cls, name: str) -> 'MXPrintMode':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


class Direction(Enum):
    PUBLISH = 'PUBLISH'
    RECEIVED = 'RECEIVED'


class PrintTag:
    # MQTT protocol
    GOOD = '[%-30s]' % colored('✔✔✔', 'green')
    DUP = '[%-30s]' % colored('DUP✔', 'green')
    ERROR = '[%-30s]' % colored('✖✖✖', 'red')

    CONNECT = '[%-30s]' % colored('-> CONNECT', 'blue')
    DISCONNECT = '[%-30s]' % colored('-> DISCONNECT', 'blue')

    SUBSCRIBE = '[%-30s]' % colored('-> SUBSCRIBE', 'white')
    UNSUBSCRIBE = '[%-30s]' % colored('-> UNSUBSCRIBE', 'white')



from big_thing_py.utils.log_util import *
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning  # type: ignore
from enum import Enum
from typing import Optional, Dict, Any

from dataclasses import dataclass

# Suppress only the single InsecureRequestWarning from urllib3 needed
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class RequestMethod(Enum):
    GET = 'GET'
    OPTIONS = 'OPTIONS'
    HEAD = 'HEAD'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'


@dataclass
class RequestResponse:
    ok: bool
    reason: str
    status_code: int
    json: dict = None
    text: str = None


def api_request(
    method: RequestMethod,
    url: str,
    query: Dict[str, Any] = None,
    data: Dict[str, Any] = None,
    headers: Dict[str, Any] = None,
    body: Dict[str, Any] = None,
    auth: Tuple[str, str] = None,
    verify: bool = False,
    timeout: float = None,
    retries: int = 3,
) -> Optional[RequestResponse]:
    headers = headers or {}
    for attempt in range(retries):
        try:
            response = requests.request(
                method=method.value,
                url=url,
                params=query,
                data=data,
                headers=headers,
                json=body,
                auth=auth,
                verify=verify,
                timeout=(1, timeout) if timeout else None,
            )

            ok = response.ok
            reason = response.reason
            status_code = response.status_code
            try:
                json_string = response.json()
            except Exception as e:
                json_string = None
            text = response.text

            try:
                request_response = RequestResponse(ok=ok, reason=reason, status_code=status_code, json=json_string, text=text)
            except Exception as e:
                request_response = RequestResponse(ok=ok, reason=reason, status_code=status_code, json=json_string, text=text)

            if ok:
                return request_response
            else:
                MXLOG_ERROR(f'Response validation failed, Status Code: {response.status_code}. Attempt: {attempt + 1}')
        except requests.exceptions.RequestException as e:
            MXLOG_WARN(f'Failed to request API: {e}. Attempt: {attempt + 1}')
    else:
        MXLOG_ERROR('All API requests failed after retries.')
        return request_response


if __name__ == '__main__':
    response = api_request('https://www.google.com')
    print(response)
